# Copyright (c) 2013-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Sam Crooks - initial implementation
#     Ron Frederick - minor cleanup

"""Logging functions"""

import logging
from typing import MutableMapping, Optional, Tuple, Union, cast


_LogArg = object
_ObjDict = MutableMapping[str, object]


class SSHLogger(logging.LoggerAdapter):
    """Adapter to add context to AsyncSSH log messages"""

    _debug_level = 1
    _pkg_logger = logging.getLogger(__package__ or 'asyncssh')

    def __init__(self, parent: logging.Logger = _pkg_logger,
                 child: str = '', context: str = ''):
        self._context = context
        self._logger = parent.getChild(child) if child else parent

        super().__init__(self._logger, {})

    def _extend_context(self, context: str) -> str:
        """Extend context provided by this logger"""

        if context:
            if self._context:
                context = self._context + ', ' + context
        else:
            context = self._context

        return context

    def get_child(self, child: str = '', context: str = '') -> 'SSHLogger':
        """Return child logger with optional added context"""

        return type(self)(self._logger, child, self._extend_context(context))

    def log(self, level: int, msg: object, *args, **kwargs) -> None:
        """Log a message to the underlying logger"""

        def _item_text(item: _LogArg) -> str:
            """Convert a list item to text"""

            if isinstance(item, bytes):
                result = item.decode('utf-8', errors='backslashreplace')

                if not result.isprintable():
                    result = repr(result)[1:-1]
            elif not isinstance(item, str):
                result = str(item)
            else:
                result = item

            return result

        def _text(arg: _LogArg) -> _LogArg:
            """Convert a log argument to text"""

            result: _LogArg

            if isinstance(arg, list):
                result = ','.join(_item_text(item) for item in arg)
            elif isinstance(arg, tuple):
                host, port = arg

                if host:
                    result = f'{host}, port {port}' if port else host
                else:
                    result = f'port {port}' if port else 'dynamic port'
            elif isinstance(arg, bytes):
                result = _item_text(arg)
            else:
                result = arg

            return result

        log_args = [_text(arg) for arg in args]

        super().log(level, msg, *log_args, **kwargs)

    def process(self, msg: str, kwargs: _ObjDict) -> Tuple[str, _ObjDict]:
        """Add context to log message"""

        extra = cast(_ObjDict, kwargs.get('extra', {}))

        context = self._extend_context(cast(str, extra.get('context')))
        context = '[' + context + '] ' if context else ''

        packet = cast(bytes, extra.get('packet'))
        pktdata = ''
        offset = 0

        while packet:
            line = f'\n  {offset:08x}:'

            for b in packet[:16]:
                line += f' {b:02x}'

            line += (62 - len(line)) * ' '

            for b in packet[:16]:
                if b < 0x20 or b >= 0x80:
                    c = '.'
                elif b == ord('%'):
                    c = '%%'
                else:
                    c = chr(b)

                line += c

            pktdata += line

            packet = packet[16:]
            offset += 16

        return context + msg + pktdata, kwargs

    @classmethod
    def set_debug_level(cls, level: int) -> None:
        """Set AsyncSSH debug log level"""

        if level < 1 or level > 3:
            raise ValueError('Debug log level must be between 1 and 3')

        cls._debug_level = level

    def debug1(self, msg: str, *args: _LogArg, **kwargs: object) -> None:
        """Write a level 1 debug log message"""

        self.log(logging.DEBUG, msg, *args, **kwargs)

    def debug2(self, msg: str, *args: _LogArg, **kwargs: object) -> None:
        """Write a level 2 debug log message"""

        if self._debug_level >= 2:
            self.log(logging.DEBUG, msg, *args, **kwargs)

    def packet(self, pktid: Optional[int], packet: bytes, msg: str,
               *args: _LogArg, **kwargs: object) -> None:
        """Write a control packet debug log message"""

        if self._debug_level >= 3:
            kwargs.setdefault('extra', {})
            extra = cast(_ObjDict, kwargs.get('extra'))

            if pktid is not None:
                extra.update(context=f'pktid={pktid}')

            extra.update(packet=packet)

            self.log(logging.DEBUG, msg, *args, **kwargs)


def set_log_level(level: Union[int, str]) -> None:
    """Set the AsyncSSH log level

       This function sets the log level of the AsyncSSH logger. It
       defaults to `'NOTSET`', meaning that it will track the debug
       level set on the root Python logger.

       For additional control over the level of debug logging, see the
       function :func:`set_debug_level` for additional information.

       :param level:
           The log level to set, as defined by the `logging` module
       :type level: `int` or `str`

    """

    logger.setLevel(level)


def set_sftp_log_level(level: Union[int, str]) -> None:
    """Set the AsyncSSH SFTP/SCP log level

       This function sets the log level of the AsyncSSH SFTP/SCP logger.
       It defaults to `'NOTSET`', meaning that it will track the debug
       level set on the main AsyncSSH logger.

       For additional control over the level of debug logging, see the
       function :func:`set_debug_level` for additional information.

       :param level:
           The log level to set, as defined by the `logging` module
       :type level: `int` or `str`

    """

    sftp_logger.setLevel(level)


def set_debug_level(level: int) -> None:
    """Set the AsyncSSH debug log level

       This function sets the level of debugging logging done by the
       AsyncSSH logger, from the following options:

           ===== ====================================
           Level Description
           ===== ====================================
           1     Minimal debug logging
           2     Full debug logging
           3     Full debug logging with packet dumps
           ===== ====================================

       The debug level defaults to level 1 (minimal debug logging).

       .. note:: For this setting to have any effect, the effective log
                 level of the AsyncSSH logger must be set to DEBUG.

       .. warning:: Extreme caution should be used when setting debug
                    level to 3, as this can expose user passwords in
                    clear text. This level should generally only be
                    needed when tracking down issues with malformed
                    or incomplete packets.

       :param level:
           The debug level to set, as defined above.
       :type level: `int`

    """

    logger.set_debug_level(level)


logger = SSHLogger()
sftp_logger = logger.get_child('sftp')



# Copyright (c) 2013-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SASLprep implementation

   This module implements the stringprep algorithm defined in RFC 3454
   and the SASLprep profile of stringprep defined in RFC 4013. This
   profile is used to normalize usernames and passwords sent in the
   SSH protocol.

"""

# The stringprep module should not be flagged as deprecated
# pylint: disable=deprecated-module
import stringprep
# pylint: enable=deprecated-module
import unicodedata

from typing import Callable, Optional, Sequence
from typing_extensions import Literal


class SASLPrepError(ValueError):
    """Invalid data provided to saslprep"""


def _check_bidi(s: str) -> None:
    """Enforce bidirectional character check from RFC 3454 (stringprep)"""

    r_and_al_cat = False
    l_cat = False

    for c in s:
        if not r_and_al_cat and stringprep.in_table_d1(c):
            r_and_al_cat = True

        if not l_cat and stringprep.in_table_d2(c):
            l_cat = True

    if r_and_al_cat and l_cat:
        raise SASLPrepError('Both RandALCat and LCat characters present')

    if r_and_al_cat and not (stringprep.in_table_d1(s[0]) and
                             stringprep.in_table_d1(s[-1])):
        raise SASLPrepError('RandALCat character not at both start and end')


def _stringprep(s: str, check_unassigned: bool,
                mapping: Optional[Callable[[str], str]],
                normalization: Literal['NFC', 'NFD', 'NFKC', 'NFKD'],
                prohibited: Sequence[Callable[[str], bool]],
                bidi: bool) -> str:
    """Implement a stringprep profile as defined in RFC 3454"""

    if check_unassigned: # pragma: no branch
        for c in s:
            if stringprep.in_table_a1(c):
                raise SASLPrepError(f'Unassigned character: {c!r}')

    if mapping: # pragma: no branch
        s = mapping(s)

    if normalization: # pragma: no branch
        s = unicodedata.normalize(normalization, s)

    if prohibited: # pragma: no branch
        for c in s:
            for lookup in prohibited:
                if lookup(c):
                    raise SASLPrepError(f'Prohibited character: {c!r}')

    if bidi: # pragma: no branch
        _check_bidi(s)

    return s


def _map_saslprep(s: str) -> str:
    """Map stringprep table B.1 to nothing and C.1.2 to ASCII space"""

    r = []

    for c in s:
        if stringprep.in_table_c12(c):
            r.append(' ')
        elif not stringprep.in_table_b1(c):
            r.append(c)

    return ''.join(r)


def saslprep(s: str) -> str:
    """Implement SASLprep profile defined in RFC 4013"""

    prohibited = (stringprep.in_table_c12, stringprep.in_table_c21_c22,
                  stringprep.in_table_c3, stringprep.in_table_c4,
                  stringprep.in_table_c5, stringprep.in_table_c6,
                  stringprep.in_table_c7, stringprep.in_table_c8,
                  stringprep.in_table_c9)

    return _stringprep(s, True, _map_saslprep, 'NFKC', prohibited, True)



# Copyright (c) 2013-2025 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SSH port forwarding handlers"""

import asyncio
import socket
from types import TracebackType
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Optional
from typing import Type, cast
from typing_extensions import Self

from .misc import ChannelOpenError, SockAddr


if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .connection import SSHConnection


SSHForwarderCoro = Callable[..., Awaitable]


class SSHForwarder(asyncio.BaseProtocol):
    """SSH port forwarding connection handler"""

    def __init__(self, peer: Optional['SSHForwarder'] = None,
                 extra: Optional[Dict[str, Any]] = None):
        self._peer = peer
        self._transport: Optional[asyncio.Transport] = None
        self._inpbuf = b''
        self._eof_received = False

        if peer:
            peer.set_peer(self)

        if extra is None:
            extra = {}

        self._extra = extra

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, _exc_type: Optional[Type[BaseException]],
                        _exc_value: Optional[BaseException],
                        _traceback: Optional[TracebackType]) -> bool:
        self.close()
        return False

    def get_extra_info(self, name: str, default: Any = None) -> Any:
        """Get additional information about the forwarder

           This method returns extra information about the forwarder.
           Currently, the only information available is the value
           ``interface`` for TUN/TAP forwarders, returning the name of the
           local TUN/TAP network interface created for this forwarder.

        """

        return self._extra.get(name, default)

    def set_peer(self, peer: 'SSHForwarder') -> None:
        """Set the peer forwarder to exchange data with"""

        self._peer = peer

    def write(self, data: bytes) -> None:
        """Write data to the transport"""

        if not self._transport:
            return # pragma: no cover

        try:
            self._transport.write(data)
        except OSError: # pragma: no cover
            pass

    def write_eof(self) -> None:
        """Write end of file to the transport"""

        if not self._transport:
            return # pragma: no cover

        try:
            self._transport.write_eof()
        except OSError: # pragma: no cover
            pass

    def was_eof_received(self) -> bool:
        """Return whether end of file has been received or not"""

        return self._eof_received

    def pause_reading(self) -> None:
        """Pause reading from the transport"""

        assert self._transport is not None
        self._transport.pause_reading()

    def resume_reading(self) -> None:
        """Resume reading on the transport"""

        assert self._transport is not None
        self._transport.resume_reading()

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Handle a newly opened connection"""

        self._transport = cast(Optional['asyncio.Transport'], transport)

        sock = cast(socket.socket, transport.get_extra_info('socket'))

        if sock and sock.family in {socket.AF_INET, socket.AF_INET6}:
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Handle an incoming connection close"""

        # pylint: disable=unused-argument

        self.close()

    def session_started(self) -> None:
        """Handle session start"""

    def data_received(self, data: bytes,
                      datatype: Optional[int] = None) -> None:
        """Handle incoming data from the transport"""

        # pylint: disable=unused-argument

        if self._peer:
            self._peer.write(data)
        else:
            self._inpbuf += data

    def eof_received(self) -> bool:
        """Handle an incoming end of file from the transport"""

        self._eof_received = True

        if self._peer:
            self._peer.write_eof()

            return not self._peer.was_eof_received()
        else:
            return True

    def pause_writing(self) -> None:
        """Pause writing by asking peer to pause reading"""

        if self._peer: # pragma: no branch
            self._peer.pause_reading()

    def resume_writing(self) -> None:
        """Resume writing by asking peer to resume reading"""

        if self._peer: # pragma: no branch
            self._peer.resume_reading()

    def close(self) -> None:
        """Close this port forwarder"""

        if self._transport:
            self._transport.close()
            self._transport = None

        if self._peer:
            peer = self._peer
            self._peer = None
            peer.close()


class SSHLocalForwarder(SSHForwarder):
    """Local forwarding connection handler"""

    def __init__(self, conn: 'SSHConnection', coro: SSHForwarderCoro):
        super().__init__()
        self._conn = conn
        self._coro = coro

    async def _forward(self, *args: object) -> None:
        """Begin local forwarding"""

        def session_factory() -> SSHForwarder:
            """Return an SSH forwarder"""

            return SSHForwarder(self)

        try:
            await self._coro(session_factory, *args)
        except ChannelOpenError as exc:
            self.connection_lost(exc)
            return

        assert self._peer is not None

        if self._inpbuf:
            self._peer.write(self._inpbuf)
            self._inpbuf = b''

        if self._eof_received:
            self._peer.write_eof()

    def forward(self, *args: object) -> None:
        """Start a task to begin local forwarding"""

        self._conn.create_task(self._forward(*args))


class SSHLocalPortForwarder(SSHLocalForwarder):
    """Local TCP port forwarding connection handler"""

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Handle a newly opened connection"""

        super().connection_made(transport)

        peername = cast(SockAddr, transport.get_extra_info('peername'))

        if peername: # pragma: no branch
            orig_host, orig_port = peername[:2]

        self.forward(orig_host, orig_port)


class SSHLocalPathForwarder(SSHLocalForwarder):
    """Local UNIX domain socket forwarding connection handler"""

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Handle a newly opened connection"""

        super().connection_made(transport)
        self.forward()



# Copyright (c) 2019-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""U2F EdDSA public key encryption handler"""

from hashlib import sha256
from typing import Optional, Tuple, cast

from .crypto import EdDSAPublicKey, ed25519_available
from .packet import Byte, String, UInt32, SSHPacket
from .public_key import KeyExportError, SSHKey, SSHOpenSSHCertificateV01
from .public_key import register_public_key_alg, register_certificate_alg
from .public_key import register_sk_alg
from .sk import SSH_SK_ED25519, SSH_SK_USER_PRESENCE_REQD, sk_enroll, sk_sign


_PrivateKeyArgs = Tuple[bytes, str, int, bytes, bytes]
_PublicKeyArgs = Tuple[bytes, str]


class _SKEd25519Key(SSHKey):
    """Handler for U2F Ed25519 public key encryption"""

    _key: EdDSAPublicKey

    algorithm = b'sk-ssh-ed25519@openssh.com'
    sig_algorithms = (algorithm,)
    all_sig_algorithms = set(sig_algorithms)
    use_executor = True

    def __init__(self, public_value: bytes, application: str,
                 flags: int = 0, key_handle: Optional[bytes] = None,
                 reserved: bytes = b''):
        super().__init__(EdDSAPublicKey.construct(b'ed25519', public_value))

        self._application = application
        self._app_hash = sha256(application.encode('utf-8')).digest()
        self._flags = flags
        self._key_handle = key_handle
        self._reserved = reserved

    def __eq__(self, other: object) -> bool:
        # This isn't protected access - both objects are _SKEd25519Key instances
        # pylint: disable=protected-access

        return (isinstance(other, type(self)) and
                self._key.public_value == other._key.public_value and
                self._application == other._application and
                self._flags == other._flags and
                self._key_handle == other._key_handle and
                self._reserved == other._reserved)

    def __hash__(self) -> int:
        return hash((self._key.public_value, self._application, self._flags,
                     self._key_handle, self._reserved))

    @classmethod
    def generate(cls, algorithm: bytes, *, # type: ignore
                 application: str = 'ssh:', user: str = 'AsyncSSH',
                 pin: Optional[str] = None, resident: bool = False,
                 touch_required: bool = True) -> '_SKEd25519Key':
        """Generate a new U2F Ed25519 private key"""

        # pylint: disable=arguments-differ

        flags = SSH_SK_USER_PRESENCE_REQD if touch_required else 0

        public_value, key_handle = sk_enroll(SSH_SK_ED25519, application,
                                             user, pin, resident)

        return cls(public_value, application, flags, key_handle, b'')

    @classmethod
    def make_private(cls, key_params: object) -> SSHKey:
        """Construct a U2F Ed25519 private key"""

        public_value, application, flags, key_handle, reserved = \
            cast(_PrivateKeyArgs, key_params)

        return cls(public_value, application, flags, key_handle, reserved)

    @classmethod
    def make_public(cls, key_params: object) -> SSHKey:
        """Construct a U2F Ed25519 public key"""

        public_value, application = cast(_PublicKeyArgs, key_params)

        return cls(public_value, application)

    @classmethod
    def decode_ssh_private(cls, packet: SSHPacket) -> _PrivateKeyArgs:
        """Decode an SSH format U2F Ed25519 private key"""

        public_value = packet.get_string()
        application = packet.get_string().decode('utf-8')
        flags = packet.get_byte()
        key_handle = packet.get_string()
        reserved = packet.get_string()

        return public_value, application, flags, key_handle, reserved

    @classmethod
    def decode_ssh_public(cls, packet: SSHPacket) -> _PublicKeyArgs:
        """Decode an SSH format U2F Ed25519 public key"""

        public_value = packet.get_string()
        application = packet.get_string().decode('utf-8')

        return public_value, application

    def encode_ssh_private(self) -> bytes:
        """Encode an SSH format U2F Ed25519 private key"""

        if self._key_handle is None:
            raise KeyExportError('Key is not private')

        return b''.join((String(self._key.public_value),
                         String(self._application), Byte(self._flags),
                         String(self._key_handle), String(self._reserved)))

    def encode_ssh_public(self) -> bytes:
        """Encode an SSH format U2F Ed25519 public key"""

        return b''.join((String(self._key.public_value),
                         String(self._application)))

    def encode_agent_cert_private(self) -> bytes:
        """Encode U2F Ed25519 certificate private key data for agent"""

        return self.encode_ssh_private()

    def sign_ssh(self, data: bytes, sig_algorithm: bytes) -> bytes:
        """Compute an SSH-encoded signature of the specified data"""

        # pylint: disable=unused-argument

        if self._key_handle is None:
            raise ValueError('Key handle needed for signing')

        flags, counter, sig, _ = sk_sign(data, self._application,
                                         self._key_handle, self._flags)

        return String(sig) + Byte(flags) + UInt32(counter)

    def verify_ssh(self, data: bytes, sig_algorithm: bytes,
                   packet: SSHPacket) -> bool:
        """Verify an SSH-encoded signature of the specified data"""

        # pylint: disable=unused-argument

        sig = packet.get_string()
        flags = packet.get_byte()
        counter = packet.get_uint32()
        packet.check_end()

        if self._touch_required and not flags & SSH_SK_USER_PRESENCE_REQD:
            return False

        return self._key.verify(self._app_hash + Byte(flags) +
                                UInt32(counter) + sha256(data).digest(), sig)


if ed25519_available: # pragma: no branch
    register_sk_alg(SSH_SK_ED25519, _SKEd25519Key)

    register_public_key_alg(b'sk-ssh-ed25519@openssh.com', _SKEd25519Key, True)

    register_certificate_alg(1, b'sk-ssh-ed25519@openssh.com',
                             b'sk-ssh-ed25519-cert-v01@openssh.com',
                             _SKEd25519Key, SSHOpenSSHCertificateV01, True)



# Copyright (c) 2016-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SSH agent support code for Windows"""

# Some of the imports below won't be found when running pylint on UNIX
# pylint: disable=import-error

import asyncio
import ctypes
import ctypes.wintypes
import errno
from typing import TYPE_CHECKING, Tuple, Union, cast

from .misc import open_file


if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .agent import AgentReader, AgentWriter


try:
    import mmapfile
    import win32api
    import win32con
    import win32ui
    _pywin32_available = True
except ImportError:
    _pywin32_available = False


_AGENT_COPYDATA_ID = 0x804e50ba
_AGENT_MAX_MSGLEN = 8192
_AGENT_NAME = 'Pageant'

_DEFAULT_OPENSSH_PATH = r'\\.\pipe\openssh-ssh-agent'


def _find_agent_window() -> 'win32ui.PyCWnd':
    """Find and return the Pageant window"""

    if _pywin32_available:
        try:
            return win32ui.FindWindow(_AGENT_NAME, _AGENT_NAME)
        except win32ui.error:
            raise OSError(errno.ENOENT, 'Agent not found') from None
    else:
        raise OSError(errno.ENOENT, 'PyWin32 not installed') from None


class _CopyDataStruct(ctypes.Structure):
    """Windows COPYDATASTRUCT argument for WM_COPYDATA message"""

    _fields_ = (('dwData', ctypes.wintypes.LPARAM),
                ('cbData', ctypes.wintypes.DWORD),
                ('lpData', ctypes.c_char_p))


class _PageantTransport:
    """Transport to connect to Pageant agent on Windows"""

    def __init__(self) -> None:
        self._mapname = f'{_AGENT_NAME}{win32api.GetCurrentThreadId():08x}'

        try:
            self._mapfile = mmapfile.mmapfile('', self._mapname,
                                              _AGENT_MAX_MSGLEN, 0, 0)
        except mmapfile.error as exc:
            raise OSError(errno.EIO, str(exc)) from None

        self._cds = _CopyDataStruct(_AGENT_COPYDATA_ID, len(self._mapname) + 1,
                                    self._mapname.encode())

        self._writing = False

    def write(self, data: bytes) -> None:
        """Write request data to Pageant agent"""

        if not self._writing:
            self._mapfile.seek(0)
            self._writing = True

        try:
            self._mapfile.write(data)
        except ValueError as exc:
            raise OSError(errno.EIO, str(exc)) from None

    async def readexactly(self, n: int) -> bytes:
        """Read response data from Pageant agent"""

        if self._writing:
            cwnd = _find_agent_window()

            if not cwnd.SendMessage(win32con.WM_COPYDATA, 0,
                                    cast(int, self._cds)):
                raise OSError(errno.EIO, 'Unable to send agent request')

            self._writing = False
            self._mapfile.seek(0)

        result = self._mapfile.read(n)

        if len(result) != n:
            raise asyncio.IncompleteReadError(result, n)

        return result

    def close(self) -> None:
        """Close the connection to Pageant"""

        if self._mapfile:
            self._mapfile.close()

    async def wait_closed(self) -> None:
        """Wait for the transport to close"""


class _W10OpenSSHTransport:
    """Transport to connect to OpenSSH agent on Windows 10"""

    def __init__(self, agent_path: str):
        self._agentfile = open_file(agent_path, 'r+b')

    async def readexactly(self, n: int) -> bytes:
        """Read response data from OpenSSH agent"""

        result = self._agentfile.read(n)

        if len(result) != n:
            raise asyncio.IncompleteReadError(result, n)

        return result

    def write(self, data: bytes) -> None:
        """Write request data to OpenSSH agent"""

        self._agentfile.write(data)

    def close(self) -> None:
        """Close the connection to OpenSSH"""

        if self._agentfile:
            self._agentfile.close()

    async def wait_closed(self) -> None:
        """Wait for the transport to close"""


async def open_agent(agent_path: str) -> Tuple['AgentReader', 'AgentWriter']:
    """Open a connection to the Pageant or Windows 10 OpenSSH agent"""

    transport: Union[None, _PageantTransport, _W10OpenSSHTransport] = None

    if not agent_path:
        try:
            _find_agent_window()
            transport = _PageantTransport()
        except OSError:
            agent_path = _DEFAULT_OPENSSH_PATH

    if not transport:
        transport = _W10OpenSSHTransport(agent_path)

    return transport, transport



# Copyright (c) 2022-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""A shim around liboqs for Streamlined NTRU Prime post-quantum encryption"""

import ctypes
import ctypes.util
from typing import Mapping, Tuple


_pq_algs: Mapping[bytes, Tuple[int, int, int, int, str]] = {
    b'mlkem768':  (1184, 2400, 1088, 32, 'KEM_ml_kem_768'),
    b'mlkem1024': (1568, 3168, 1568, 32, 'KEM_ml_kem_1024'),
    b'sntrup761': (1158, 1763, 1039, 32, 'KEM_ntruprime_sntrup761')
}

mlkem_available = False
sntrup_available = False

for lib in ('oqs', 'liboqs'):
    _oqs_lib = ctypes.util.find_library(lib)

    if _oqs_lib: # pragma: no branch
        break
else: # pragma: no cover
    _oqs_lib = None

if _oqs_lib: # pragma: no branch
    _oqs = ctypes.cdll.LoadLibrary(_oqs_lib)

    mlkem_available = (hasattr(_oqs, 'OQS_KEM_ml_kem_768_keypair') or
                       hasattr(_oqs, 'OQS_KEM_ml_kem_768_ipd_keypair'))
    sntrup_available = hasattr(_oqs, 'OQS_KEM_ntruprime_sntrup761_keypair')


class PQDH:
    """A shim around liboqs for post-quantum key exchange algorithms"""

    def __init__(self, alg_name: bytes):
        try:
            self.pubkey_bytes, self.privkey_bytes, \
            self.ciphertext_bytes, self.secret_bytes, \
            oqs_name = _pq_algs[alg_name]
        except KeyError:
            raise ValueError('Unknown PQ algorithm ' +
                             alg_name.decode()) from None

        if not hasattr(_oqs, 'OQS_' + oqs_name + # pragma: no cover
                       '_keypair'):
            oqs_name += '_ipd'

        self._keypair = getattr(_oqs, 'OQS_' + oqs_name + '_keypair')
        self._encaps = getattr(_oqs, 'OQS_' + oqs_name + '_encaps')
        self._decaps = getattr(_oqs, 'OQS_' + oqs_name + '_decaps')

    def keypair(self) -> Tuple[bytes, bytes]:
        """Make a new key pair"""

        pubkey = ctypes.create_string_buffer(self.pubkey_bytes)
        privkey = ctypes.create_string_buffer(self.privkey_bytes)
        self._keypair(pubkey, privkey)

        return pubkey.raw, privkey.raw

    def encaps(self, pubkey: bytes) -> Tuple[bytes, bytes]:
        """Generate a random secret and encrypt it with a public key"""

        if len(pubkey) != self.pubkey_bytes:
            raise ValueError('Invalid public key')

        ciphertext = ctypes.create_string_buffer(self.ciphertext_bytes)
        secret = ctypes.create_string_buffer(self.secret_bytes)

        self._encaps(ciphertext, secret, pubkey)

        return secret.raw, ciphertext.raw

    def decaps(self, ciphertext: bytes, privkey: bytes) -> bytes:
        """Decrypt an encrypted secret using a private key"""

        if len(ciphertext) != self.ciphertext_bytes:
            raise ValueError('Invalid ciphertext')

        secret = ctypes.create_string_buffer(self.secret_bytes)

        self._decaps(secret, ciphertext, privkey)

        return secret.raw



# Copyright (c) 2016-2025 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SSH process handlers"""

import asyncio
from asyncio.subprocess import DEVNULL, PIPE, STDOUT
import codecs
import inspect
import io
import os
from pathlib import PurePath
import socket
import stat
from types import TracebackType
from typing import Any, AnyStr, Awaitable, Callable, Dict, Generic, IO
from typing import Iterable, List, Mapping, Optional, Set, TextIO
from typing import Tuple, Type, TypeVar, Union, cast
from typing_extensions import Protocol, Self

from .channel import SSHChannel, SSHClientChannel, SSHServerChannel

from .constants import DEFAULT_LANG, EXTENDED_DATA_STDERR

from .logging import SSHLogger

from .misc import BytesOrStr, Error, MaybeAwait, TermModes, TermSize
from .misc import ProtocolError, Record, open_file, set_terminal_size
from .misc import BreakReceived, SignalReceived, TerminalSizeChanged

from .session import DataType

from .stream import SSHReader, SSHWriter, SSHStreamSession
from .stream import SSHClientStreamSession, SSHServerStreamSession
from .stream import SFTPServerFactory

_AnyStrContra = TypeVar('_AnyStrContra', bytes, str, contravariant=True)

_File = Union[IO[bytes], '_AsyncFileProtocol[bytes]']

ProcessSource = Union[int, str, socket.socket, PurePath, SSHReader[bytes],
                      asyncio.StreamReader, _File]

ProcessTarget = Union[int, str, socket.socket, PurePath, SSHWriter[bytes],
                      asyncio.StreamWriter, _File]

SSHServerProcessFactory = Callable[['SSHServerProcess[AnyStr]'],
                                   MaybeAwait[None]]


_QUEUE_LOW_WATER = 8
_QUEUE_HIGH_WATER = 16


class _AsyncFileProtocol(Protocol[AnyStr]):
    """Protocol for an async file"""

    async def read(self, n: int = -1) -> AnyStr:
        """Read from an async file"""

    async def write(self, data: AnyStr) -> None:
        """Write to an async file"""

    async def close(self) -> None:
        """Close an async file"""


class _ReaderProtocol(Protocol):
    """A class that can be used as a reader in SSHProcess"""

    def pause_reading(self) -> None:
        """Pause reading"""

    def resume_reading(self) -> None:
        """Resume reading"""

    def close(self) -> None:
        """Stop forwarding data"""


class _WriterProtocol(Protocol[_AnyStrContra]):
    """A class that can be used as a writer in SSHProcess"""

    def write(self, data: _AnyStrContra) -> None:
        """Write data"""

    def write_exception(self, exc: Exception) -> None:
        """Write exception (break, signal, terminal size change)"""

        return # pragma: no cover

    def write_eof(self) -> None:
        """Close output when end of file is received"""

    def close(self) -> None:
        """Stop forwarding data"""


def _is_regular_file(file: IO[bytes]) -> bool:
    """Return if argument is a regular file or file-like object"""

    try:
        return stat.S_ISREG(os.fstat(file.fileno()).st_mode)
    except OSError:
        return True

class _UnicodeReader(_ReaderProtocol, Generic[AnyStr]):
    """Handle buffering partial Unicode data"""

    def __init__(self, encoding: Optional[str], errors: str,
                 textmode: bool = False):
        super().__init__()

        if encoding and not textmode:
            self._decoder: Optional[codecs.IncrementalDecoder] = \
                codecs.getincrementaldecoder(encoding)(errors)
        else:
            self._decoder = None

    def decode(self, data: bytes, final: bool = False) -> AnyStr:
        """Decode Unicode bytes when reading from binary sources"""

        if self._decoder:
            try:
                decoded_data = cast(AnyStr, self._decoder.decode(data, final))
            except UnicodeDecodeError as exc:
                raise ProtocolError(str(exc)) from None
        else:
            decoded_data = cast(AnyStr, data)

        return decoded_data

    def check_partial(self) -> None:
        """Check if there's partial Unicode data left at EOF"""

        self.decode(b'', True)

    def close(self) -> None:
        """Perform necessary cleanup on error (provided by derived classes)"""


class _UnicodeWriter(_WriterProtocol[AnyStr]):
    """Handle encoding Unicode data before writing it"""

    def __init__(self, encoding: Optional[str], errors: str,
                 textmode: bool = False):
        super().__init__()

        if encoding and not textmode:
            self._encoder: Optional[codecs.IncrementalEncoder] = \
                codecs.getincrementalencoder(encoding)(errors)
        else:
            self._encoder = None

    def encode(self, data: AnyStr) -> bytes:
        """Encode Unicode bytes when writing to binary targets"""

        if self._encoder:
            assert self._encoder is not None
            encoded_data = cast(bytes, self._encoder.encode(cast(str, data)))
        else:
            encoded_data = cast(bytes, data)

        return encoded_data


class _FileReader(_UnicodeReader[AnyStr]):
    """Forward data from a file"""

    def __init__(self, process: 'SSHProcess[AnyStr]', file: IO[bytes],
                 bufsize: int, datatype: DataType,
                 encoding: Optional[str], errors: str):
        super().__init__(encoding, errors, hasattr(file, 'encoding'))

        self._process: 'SSHProcess[AnyStr]' = process
        self._file = file
        self._bufsize = bufsize
        self._datatype = datatype
        self._paused = False

    def feed(self) -> None:
        """Feed file data"""

        while not self._paused:
            data = self._file.read(self._bufsize)

            if data:
                self._process.feed_data(self.decode(data), self._datatype)
            else:
                self.check_partial()
                self._process.feed_eof(self._datatype)
                break

    def pause_reading(self) -> None:
        """Pause reading from the file"""

        self._paused = True

    def resume_reading(self) -> None:
        """Resume reading from the file"""

        self._paused = False
        self.feed()

    def close(self) -> None:
        """Stop forwarding data from the file"""

        self._file.close()


class _AsyncFileReader(_UnicodeReader[AnyStr]):
    """Forward data from an aiofile"""

    def __init__(self, process: 'SSHProcess[AnyStr]',
                 file: _AsyncFileProtocol[bytes],
                 bufsize: int, datatype: DataType,
                 encoding: Optional[str], errors: str):
        super().__init__(encoding, errors, hasattr(file, 'encoding'))

        self._conn = process.channel.get_connection()
        self._process: 'SSHProcess[AnyStr]' = process
        self._file = file
        self._bufsize = bufsize
        self._datatype = datatype
        self._paused = False

    async def _feed(self) -> None:
        """Feed file data"""

        while not self._paused:
            data = await self._file.read(self._bufsize)

            if data:
                self._process.feed_data(self.decode(data), self._datatype)
            else:
                self.check_partial()
                self._process.feed_eof(self._datatype)
                break

    def feed(self) -> None:
        """Start feeding file data"""

        self._conn.create_task(self._feed())

    def pause_reading(self) -> None:
        """Pause reading from the file"""

        self._paused = True

    def resume_reading(self) -> None:
        """Resume reading from the file"""

        self._paused = False
        self.feed()

    def close(self) -> None:
        """Stop forwarding data from the file"""

        self._conn.create_task(self._file.close())


class _FileWriter(_UnicodeWriter[AnyStr]):
    """Forward data to a file"""

    def __init__(self, file: IO[bytes], needs_close: bool,
                 encoding: Optional[str], errors: str):
        super().__init__(encoding, errors, hasattr(file, 'encoding'))

        self._file = file
        self._needs_close = needs_close

    def write(self, data: AnyStr) -> None:
        """Write data to the file"""

        self._file.write(self.encode(data))

    def write_eof(self) -> None:
        """Close output file when end of file is received"""

        self.close()

    def close(self) -> None:
        """Stop forwarding data to the file"""

        if self._needs_close:
            self._file.close()


class _AsyncFileWriter(_UnicodeWriter[AnyStr]):
    """Forward data to an aiofile"""

    def __init__(self, process: 'SSHProcess[AnyStr]',
                 file: _AsyncFileProtocol[bytes], needs_close: bool,
                 datatype: Optional[int], encoding: Optional[str], errors: str):
        super().__init__(encoding, errors, hasattr(file, 'encoding'))

        self._process: 'SSHProcess[AnyStr]' = process
        self._file = file
        self._needs_close = needs_close
        self._datatype = datatype
        self._paused = False
        self._queue: asyncio.Queue[Optional[AnyStr]] = asyncio.Queue()
        self._write_task: Optional[asyncio.Task[None]] = \
            process.channel.get_connection().create_task(self._writer())

    async def _writer(self) -> None:
        """Process writes to the file"""

        while True:
            data = await self._queue.get()

            if data is None:
                self._queue.task_done()
                break

            await self._file.write(self.encode(data))
            self._queue.task_done()

            if self._paused and self._queue.qsize() < _QUEUE_LOW_WATER:
                self._process.resume_feeding(self._datatype)
                self._paused = False

        if self._needs_close:
            await self._file.close()

    def write(self, data: AnyStr) -> None:
        """Write data to the file"""

        self._queue.put_nowait(data)

        if not self._paused and self._queue.qsize() >= _QUEUE_HIGH_WATER:
            self._paused = True
            self._process.pause_feeding(self._datatype)

    def write_eof(self) -> None:
        """Close output file when end of file is received"""

        self.close()

    def close(self) -> None:
        """Stop forwarding data to the file"""

        if self._write_task:
            self._write_task = None
            self._queue.put_nowait(None)
            self._process.add_cleanup_task(self._queue.join())


class _PipeReader(_UnicodeReader[AnyStr], asyncio.BaseProtocol):
    """Forward data from a pipe"""

    def __init__(self, process: 'SSHProcess[AnyStr]', datatype: DataType,
                 encoding: Optional[str], errors: str):
        super().__init__(encoding, errors)

        self._process: 'SSHProcess[AnyStr]' = process
        self._datatype = datatype
        self._transport: Optional[asyncio.ReadTransport] = None

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Handle a newly opened pipe"""

        self._transport = cast(asyncio.ReadTransport, transport)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Handle closing of the pipe"""

        self._process.feed_close(self._datatype)
        self.close()

    def data_received(self, data: bytes) -> None:
        """Forward data from the pipe"""

        self._process.feed_data(self.decode(data), self._datatype)

    def eof_received(self) -> None:
        """Forward EOF from the pipe"""

        self.check_partial()
        self._process.feed_eof(self._datatype)

    def pause_reading(self) -> None:
        """Pause reading from the pipe"""

        assert self._transport is not None
        self._transport.pause_reading()

    def resume_reading(self) -> None:
        """Resume reading from the pipe"""

        assert self._transport is not None
        self._transport.resume_reading()

    def close(self) -> None:
        """Stop forwarding data from the pipe"""

        assert self._transport is not None
        self._transport.close()


class _PipeWriter(_UnicodeWriter[AnyStr], asyncio.BaseProtocol):
    """Forward data to a pipe"""

    def __init__(self, process: 'SSHProcess[AnyStr]', datatype: DataType,
                 encoding: Optional[str], errors: str):
        super().__init__(encoding, errors)

        self._process: 'SSHProcess[AnyStr]' = process
        self._datatype = datatype
        self._transport: Optional[asyncio.WriteTransport] = None
        self._tty: Optional[IO] = None
        self._close_event = asyncio.Event()

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Handle a newly opened pipe"""

        self._transport = cast(asyncio.WriteTransport, transport)

        pipe = transport.get_extra_info('pipe')

        if isinstance(self._process, SSHServerProcess) and pipe.isatty():
            self._tty = pipe
            set_terminal_size(pipe, *self._process.term_size)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Handle closing of the pipe"""

        self._close_event.set()

    def pause_writing(self) -> None:
        """Pause writing to the pipe"""

        self._process.pause_feeding(self._datatype)

    def resume_writing(self) -> None:
        """Resume writing to the pipe"""

        self._process.resume_feeding(self._datatype)

    def write(self, data: AnyStr) -> None:
        """Write data to the pipe"""

        assert self._transport is not None
        self._transport.write(self.encode(data))

    def write_exception(self, exc: Exception) -> None:
        """Write terminal size changes to the pipe if it is a TTY"""

        if isinstance(exc, TerminalSizeChanged) and self._tty:
            set_terminal_size(self._tty, *exc.term_size)

    def write_eof(self) -> None:
        """Write EOF to the pipe"""

        assert self._transport is not None
        self._transport.write_eof()

    def close(self) -> None:
        """Stop forwarding data to the pipe"""

        assert self._transport is not None
        self._transport.close()
        self._process.add_cleanup_task(self._close_event.wait())


class _ProcessReader(_ReaderProtocol, Generic[AnyStr]):
    """Forward data from another SSH process"""

    def __init__(self, process: 'SSHProcess[AnyStr]', datatype: DataType):
        super().__init__()
        self._process: 'SSHProcess[AnyStr]' = process
        self._datatype = datatype

    def pause_reading(self) -> None:
        """Pause reading from the other channel"""

        self._process.pause_feeding(self._datatype)

    def resume_reading(self) -> None:
        """Resume reading from the other channel"""

        self._process.resume_feeding(self._datatype)

    def close(self) -> None:
        """Stop forwarding data from the other channel"""

        self._process.clear_writer(self._datatype)


class _ProcessWriter(_WriterProtocol[AnyStr]):
    """Forward data to another SSH process"""

    def __init__(self, process: 'SSHProcess[AnyStr]', datatype: DataType):
        super().__init__()
        self._process: 'SSHProcess[AnyStr]' = process
        self._datatype = datatype

    def write(self, data: AnyStr) -> None:
        """Write data to the other channel"""

        self._process.feed_data(data, self._datatype)

    def write_exception(self, exc: Exception) -> None:
        """Write an exception to the other channel"""

        cast(SSHClientProcess, self._process).feed_exception(exc)

    def write_eof(self) -> None:
        """Write EOF to the other channel"""

        self._process.feed_eof(self._datatype)

    def close(self) -> None:
        """Stop forwarding data to the other channel"""

        self._process.clear_reader(self._datatype)


class _StreamReader(_UnicodeReader[AnyStr]):
    """Forward data from an asyncio stream"""

    def __init__(self, process: 'SSHProcess[AnyStr]',
                 reader: asyncio.StreamReader,
                 bufsize: int, datatype: DataType,
                 encoding: Optional[str], errors: str):
        super().__init__(encoding, errors)

        self._process: 'SSHProcess[AnyStr]' = process
        self._conn = process.channel.get_connection()
        self._reader = reader
        self._bufsize = bufsize
        self._datatype = datatype
        self._paused = False

    async def _feed(self) -> None:
        """Feed stream data"""

        while not self._paused:
            data = await self._reader.read(self._bufsize)

            if data:
                self._process.feed_data(self.decode(data), self._datatype)
            else:
                self.check_partial()
                self._process.feed_eof(self._datatype)
                break

    def feed(self) -> None:
        """Start feeding stream data"""

        self._conn.create_task(self._feed())

    def pause_reading(self) -> None:
        """Pause reading from the stream"""

        self._paused = True

    def resume_reading(self) -> None:
        """Resume reading from the stream"""

        self._paused = False
        self.feed()

    def close(self) -> None:
        """Ignore close -- the caller must clean up the associated transport"""


class _StreamWriter(_UnicodeWriter[AnyStr]):
    """Forward data to an asyncio stream"""

    def __init__(self, process: 'SSHProcess[AnyStr]',
                 writer: asyncio.StreamWriter, recv_eof: bool,
                 datatype: Optional[int], encoding: Optional[str], errors: str):
        super().__init__(encoding, errors)

        self._process: 'SSHProcess[AnyStr]' = process
        self._writer = writer
        self._recv_eof = recv_eof
        self._datatype = datatype
        self._paused = False
        self._queue: asyncio.Queue[Optional[AnyStr]] = asyncio.Queue()
        self._write_task: Optional[asyncio.Task[None]] = \
            process.channel.get_connection().create_task(self._feed())

    async def _feed(self) -> None:
        """Feed data to the stream"""

        while True:
            data = await self._queue.get()

            if data is None:
                self._queue.task_done()
                break

            self._writer.write(self.encode(data))
            await self._writer.drain()
            self._queue.task_done()

            if self._paused and self._queue.qsize() < _QUEUE_LOW_WATER:
                self._process.resume_feeding(self._datatype)
                self._paused = False

        if self._recv_eof:
            self._writer.write_eof()

    def write(self, data: AnyStr) -> None:
        """Write data to the stream"""

        self._queue.put_nowait(data)

        if not self._paused and self._queue.qsize() >= _QUEUE_HIGH_WATER:
            self._paused = True
            self._process.pause_feeding(self._datatype)

    def write_eof(self) -> None:
        """Write EOF to the stream"""

        self.close()

    def close(self) -> None:
        """Stop forwarding data to the stream"""

        if self._write_task:
            self._write_task = None
            self._queue.put_nowait(None)
            self._process.add_cleanup_task(self._queue.join())


class _DevNullWriter(_WriterProtocol[AnyStr]):
    """Discard data"""

    def write(self, data: AnyStr) -> None:
        """Discard data being written"""

    def write_eof(self) -> None:
        """Ignore end of file"""

    def close(self) -> None:
        """Ignore close"""


class _StdoutWriter(_WriterProtocol[AnyStr]):
    """Forward data to an SSH process' stdout instead of stderr"""

    def __init__(self, process: 'SSHProcess[AnyStr]'):
        super().__init__()
        self._process: 'SSHProcess[AnyStr]' = process

    def write(self, data: AnyStr) -> None:
        """Pretend data was received on stdout"""

        self._process.data_received(data, None)

    def write_eof(self) -> None:
        """Ignore end of file"""

    def close(self) -> None:
        """Ignore close"""


class ProcessError(Error):
    """SSH Process error

       This exception is raised when an :class:`SSHClientProcess` exits
       with a non-zero exit status and error checking is enabled. In
       addition to the usual error code, reason, and language, it
       contains the following fields:

         ============ ======================================= =================
         Field        Description                             Type
         ============ ======================================= =================
         env          The environment the client requested    `str` or `None`
                      to be set for the process
         command      The command the client requested the    `str` or `None`
                      process to execute (if any)
         subsystem    The subsystem the client requested the  `str` or `None`
                      process to open (if any)
         exit_status  The exit status returned, or -1 if an   `int` or `None`
                      exit signal is sent
         exit_signal  The exit signal sent (if any) in the    `tuple` or `None`
                      form of a tuple containing the signal
                      name, a `bool` for whether a core dump
                      occurred, a message associated with the
                      signal, and the language the message
                      was in
         returncode   The exit status returned, or negative   `int` or `None`
                      of the signal number when an exit
                      signal is sent
         stdout       The output sent by the process to       `str` or `bytes`
                      stdout (if not redirected)
         stderr       The output sent by the process to       `str` or `bytes`
                      stderr (if not redirected)
         ============ ======================================= =================

    """

    def __init__(self, env: Optional[Mapping[str, str]],
                 command: Optional[str], subsystem: Optional[str],
                 exit_status: Optional[int],
                 exit_signal: Optional[Tuple[str, bool, str, str]],
                 returncode: Optional[int], stdout: BytesOrStr,
                 stderr: BytesOrStr, reason: str = '',
                 lang: str = DEFAULT_LANG):
        self.env = env
        self.command = command
        self.subsystem = subsystem
        self.exit_status = exit_status
        self.exit_signal = exit_signal
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr

        if exit_signal:
            signal, core_dumped, msg, lang = exit_signal
            reason = 'Process exited with signal ' + signal + \
                     (': ' + msg if msg else '') + \
                     (' (core dumped)' if core_dumped else '')
        elif exit_status:
            reason = f'Process exited with non-zero exit status {exit_status}'

        super().__init__(exit_status or 0, reason, lang)


# pylint: disable=redefined-builtin
class TimeoutError(ProcessError, asyncio.TimeoutError):
    """SSH Process timeout error

       This exception is raised when a timeout occurs when calling the
       :meth:`wait <SSHClientProcess.wait>` method on :class:`SSHClientProcess`
       or the :meth:`run <SSHClientConnection.run>` method on
       :class:`SSHClientConnection`. It is a subclass of :class:`ProcessError`
       and contains all of the fields documented there, including any output
       received on stdout and stderr prior to when the timeout occurred. It
       is also a subclass of :class:`asyncio.TimeoutError`, for code that
       might be expecting that.

    """
# pylint: enable=redefined-builtin


class SSHCompletedProcess(Record):
    """Results from running an SSH process

       This object is returned by the :meth:`run <SSHClientConnection.run>`
       method on :class:`SSHClientConnection` when the requested command
       has finished running. It contains the following fields:

         ============ ======================================= =================
         Field        Description                             Type
         ============ ======================================= =================
         env          The environment the client requested    `dict` or `None`
                      to be set for the process
         command      The command the client requested the    `str` or `None`
                      process to execute (if any)
         subsystem    The subsystem the client requested the  `str` or `None`
                      process to open (if any)
         exit_status  The exit status returned, or -1 if an   `int`
                      exit signal is sent
         exit_signal  The exit signal sent (if any) in the    `tuple` or `None`
                      form of a tuple containing the signal
                      name, a `bool` for whether a core dump
                      occurred, a message associated with the
                      signal, and the language the message
                      was in
         returncode   The exit status returned, or negative   `int`
                      of the signal number when an exit
                      signal is sent
         stdout       The output sent by the process to       `str` or `bytes`
                      stdout (if not redirected)
         stderr       The output sent by the process to       `str` or `bytes`
                      stderr (if not redirected)
         ============ ======================================= =================

    """

    env: Optional[Mapping[str, str]]
    command: Optional[str]
    subsystem: Optional[str]
    exit_status: Optional[int]
    exit_signal: Optional[Tuple[str, bool, str, str]]
    returncode: Optional[int]
    stdout: Optional[BytesOrStr]
    stderr: Optional[BytesOrStr]


class SSHProcess(SSHStreamSession, Generic[AnyStr]):
    """SSH process handler"""

    def __init__(self, *args) -> None:
        super().__init__(*args)

        self._cleanup_tasks: List[Awaitable[None]] = []

        self._readers: Dict[Optional[int], _ReaderProtocol] = {}
        self._send_eof: Dict[Optional[int], bool] = {}

        self._writers: Dict[Optional[int], _WriterProtocol[AnyStr]] = {}
        self._recv_eof: Dict[Optional[int], bool] = {}

        self._paused_write_streams: Set[Optional[int]] = set()

    async def __aenter__(self) -> Self:
        """Allow SSHProcess to be used as an async context manager"""

        return self

    async def __aexit__(self, _exc_type: Optional[Type[BaseException]],
                        _exc_value: Optional[BaseException],
                        _traceback: Optional[TracebackType]) -> bool:
        """Wait for a full channel close when exiting the async context"""

        self.close()
        await self.wait_closed()
        return False

    @property
    def channel(self) -> SSHChannel[AnyStr]:
        """The channel associated with the process"""

        assert self._chan is not None
        return self._chan

    @property
    def logger(self) -> SSHLogger:
        """The logger associated with the process"""

        assert self._chan is not None
        return self._chan.logger

    @property
    def command(self) -> Optional[str]:
        """The command the client requested to execute, if any

           If the client did not request that a command be executed,
           this property will be set to `None`.

        """

        assert self._chan is not None
        return self._chan.get_command()

    @property
    def subsystem(self) -> Optional[str]:
        """The subsystem the client requested to open, if any

           If the client did not request that a subsystem be opened,
           this property will be set to `None`.

        """

        assert self._chan is not None
        return self._chan.get_subsystem()

    @property
    def env(self) -> Mapping[str, str]:
        """A mapping containing the environment set by the client"""

        assert self._chan is not None
        return self._chan.get_environment()

    def get_extra_info(self, name: str, default: Any = None) -> Any:
        """Return additional information about this process

           This method returns extra information about the channel
           associated with this process. See :meth:`get_extra_info()
           <SSHClientChannel.get_extra_info>` on :class:`SSHClientChannel`
           for additional information.

        """

        assert self._chan is not None
        return self._chan.get_extra_info(name, default)

    async def _create_reader(self, source: ProcessSource, bufsize: int,
                             send_eof: bool, recv_eof: bool,
                             datatype: DataType = None) -> None:
        """Create a reader to forward data to the SSH channel"""

        def pipe_factory() -> _PipeReader:
            """Return a pipe read handler"""

            return _PipeReader(self, datatype, self._encoding, self._errors)

        if source == PIPE:
            reader: Optional[_ReaderProtocol] = None
        elif source == DEVNULL:
            assert self._chan is not None
            self._chan.write_eof()
            reader = None
        elif isinstance(source, SSHReader):
            reader_stream, reader_datatype = source.get_redirect_info()
            reader_process = cast('SSHProcess[AnyStr]', reader_stream)
            writer = _ProcessWriter[AnyStr](self, datatype)
            reader_process.set_writer(writer, recv_eof, reader_datatype)
            reader = _ProcessReader(reader_process, reader_datatype)
        elif isinstance(source, asyncio.StreamReader):
            reader = _StreamReader(self, source, bufsize, datatype,
                                   self._encoding, self._errors)
        else:
            file: _File

            if isinstance(source, str):
                file = open_file(source, 'rb', buffering=bufsize)
            elif isinstance(source, PurePath):
                file = open_file(str(source), 'rb', buffering=bufsize)
            elif isinstance(source, int):
                file = os.fdopen(source, 'rb', buffering=bufsize)
            elif isinstance(source, socket.socket):
                file = os.fdopen(source.detach(), 'rb', buffering=bufsize)
            else:
                file = source

            if hasattr(file, 'read') and \
                    (inspect.iscoroutinefunction(file.read) or
                     inspect.isgeneratorfunction(file.read)):
                reader = _AsyncFileReader(self, cast(_AsyncFileProtocol, file),
                                          bufsize, datatype, self._encoding,
                                          self._errors)
            elif _is_regular_file(cast(IO[bytes], file)):
                reader = _FileReader(self, cast(IO[bytes], file), bufsize,
                                     datatype, self._encoding, self._errors)
            else:
                if hasattr(source, 'buffer'):
                    # If file was opened in text mode, remove that wrapper
                    file = cast(TextIO, source).buffer

                assert self._loop is not None
                _, protocol = \
                    await self._loop.connect_read_pipe(pipe_factory, file)
                reader = cast(_PipeReader, protocol)

        self.set_reader(reader, send_eof, datatype)

        if isinstance(reader, (_FileReader, _AsyncFileReader, _StreamReader)):
            reader.feed()
        elif isinstance(reader, _ProcessReader):
            reader_process.feed_recv_buf(reader_datatype, writer)

    async def _create_writer(self, target: ProcessTarget, bufsize: int,
                             send_eof: bool, recv_eof: bool,
                             datatype: DataType = None) -> None:
        """Create a writer to forward data from the SSH channel"""

        def pipe_factory() -> _PipeWriter:
            """Return a pipe write handler"""

            return _PipeWriter(self, datatype, self._encoding, self._errors)

        if target == PIPE:
            writer: Optional[_WriterProtocol[AnyStr]] = None
        elif target == DEVNULL:
            writer = _DevNullWriter()
        elif target == STDOUT:
            writer = _StdoutWriter(self)
        elif isinstance(target, SSHWriter):
            writer_stream, writer_datatype = target.get_redirect_info()
            writer_process = cast('SSHProcess[AnyStr]', writer_stream)
            reader = _ProcessReader(self, datatype)
            writer_process.set_reader(reader, send_eof, writer_datatype)
            writer = _ProcessWriter[AnyStr](writer_process, writer_datatype)
        elif isinstance(target, asyncio.StreamWriter):
            writer = _StreamWriter(self, target, recv_eof, datatype,
                                   self._encoding, self._errors)
        else:
            file: _File
            needs_close = True

            if isinstance(target, str):
                file = open_file(target, 'wb', buffering=bufsize)
            elif isinstance(target, PurePath):
                file = open_file(str(target), 'wb', buffering=bufsize)
            elif isinstance(target, int):
                file = os.fdopen(target, 'wb',
                                 buffering=bufsize, closefd=recv_eof)
            elif isinstance(target, socket.socket):
                fd = target.detach() if recv_eof else target.fileno()
                file = os.fdopen(fd, 'wb', buffering=bufsize, closefd=recv_eof)
            else:
                file = target
                needs_close = recv_eof

            if hasattr(file, 'write') and \
                    (inspect.iscoroutinefunction(file.write) or
                     inspect.isgeneratorfunction(file.write)):
                writer = _AsyncFileWriter(
                    self, cast(_AsyncFileProtocol, file), needs_close,
                    datatype, self._encoding, self._errors)
            elif _is_regular_file(cast(IO[bytes], file)):
                writer = _FileWriter(cast(IO[bytes], file), needs_close,
                                    self._encoding, self._errors)
            else:
                if hasattr(target, 'buffer'):
                    # If file was opened in text mode, remove that wrapper
                    file = cast(TextIO, target).buffer

                if not recv_eof:
                    fd = os.dup(cast(IO[bytes], file).fileno())
                    file = os.fdopen(fd, 'wb', buffering=0)

                assert self._loop is not None
                _, protocol = \
                    await self._loop.connect_write_pipe(pipe_factory, file)
                writer = cast(_PipeWriter, protocol)

        self.set_writer(writer, recv_eof, datatype)

        if writer:
            self.feed_recv_buf(datatype, writer)

    def _should_block_drain(self, datatype: DataType) -> bool:
        """Return whether output is still being written to the channel"""

        return (datatype in self._readers or
                super()._should_block_drain(datatype))

    def _should_pause_reading(self) -> bool:
        """Return whether to pause reading from the channel"""

        return bool(self._paused_write_streams) or \
            super()._should_pause_reading()

    def add_cleanup_task(self, task: Awaitable) -> None:
        """Add a task to run when the process exits"""

        self._cleanup_tasks.append(task)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Handle a close of the SSH channel"""

        super().connection_lost(exc) # type: ignore

        for reader in list(self._readers.values()):
            reader.close()

        for writer in list(self._writers.values()):
            writer.close()

        self._readers = {}
        self._writers = {}

    def data_received(self, data: AnyStr, datatype: DataType) -> None:
        """Handle incoming data from the SSH channel"""

        writer = self._writers.get(datatype)

        if writer:
            writer.write(data)
        else:
            super().data_received(data, datatype)

    def eof_received(self) -> bool:
        """Handle an incoming end of file from the SSH channel"""

        for datatype, writer in list(self._writers.items()):
            if self._recv_eof[datatype]:
                writer.write_eof()

        return super().eof_received()

    def pause_writing(self) -> None:
        """Pause forwarding data to the channel"""

        super().pause_writing()

        for reader in list(self._readers.values()):
            reader.pause_reading()

    def resume_writing(self) -> None:
        """Resume forwarding data to the channel"""

        super().resume_writing()

        for reader in list(self._readers.values()):
            reader.resume_reading()

    def feed_data(self, data: AnyStr, datatype: DataType) -> None:
        """Feed data to the channel"""

        assert self._chan is not None
        self._chan.write(data, datatype)

    def feed_eof(self, datatype: DataType) -> None:
        """Feed EOF to the channel"""

        if self._send_eof[datatype]:
            assert self._chan is not None
            self._chan.write_eof()

        self._readers[datatype].close()
        self.clear_reader(datatype)

    def feed_close(self, datatype: DataType) -> None:
        """Feed pipe close to the channel"""

        if datatype in self._readers:
            self.feed_eof(datatype)

    def feed_recv_buf(self, datatype: DataType,
                      writer: _WriterProtocol[AnyStr]) -> None:
        """Feed current receive buffer to a newly set writer"""

        for buf in self._recv_buf[datatype]:
            if isinstance(buf, Exception):
                writer.write_exception(buf)
            else:
                writer.write(buf)
                self._recv_buf_len -= len(buf)

        self._recv_buf[datatype].clear()

        if self._eof_received:
            writer.write_eof()

        self._maybe_resume_reading()

    def pause_feeding(self, datatype: DataType) -> None:
        """Pause feeding data from the channel"""

        self._paused_write_streams.add(datatype)
        self._maybe_pause_reading()

    def resume_feeding(self, datatype: DataType) -> None:
        """Resume feeding data from the channel"""

        self._paused_write_streams.remove(datatype)
        self._maybe_resume_reading()

    def set_reader(self, reader: Optional[_ReaderProtocol],
                   send_eof: bool, datatype: DataType) -> None:
        """Set a reader used to forward data to the channel"""

        old_reader = self._readers.get(datatype)

        if old_reader:
            old_reader.close()

        if reader:
            self._readers[datatype] = reader
            self._send_eof[datatype] = send_eof

            if self._write_paused:
                reader.pause_reading()
        elif old_reader:
            self.clear_reader(datatype)

    def clear_reader(self, datatype: DataType) -> None:
        """Clear a reader forwarding data to the channel"""

        del self._readers[datatype]
        del self._send_eof[datatype]
        self._unblock_drain(datatype)

    def set_writer(self, writer: Optional[_WriterProtocol[AnyStr]],
                   recv_eof: bool, datatype: DataType) -> None:
        """Set a writer used to forward data from the channel"""

        old_writer = self._writers.get(datatype)

        if old_writer:
            old_writer.close()
            self.clear_writer(datatype)

        if writer:
            self._writers[datatype] = writer
            self._recv_eof[datatype] = recv_eof

    def clear_writer(self, datatype: DataType) -> None:
        """Clear a writer forwarding data from the channel"""

        if datatype in self._paused_write_streams:
            self.resume_feeding(datatype)

        del self._writers[datatype]

    def close(self) -> None:
        """Shut down the process"""

        assert self._chan is not None
        self._chan.close()

    def is_closing(self) -> bool:
        """Return if the channel is closing or is closed"""

        assert self._chan is not None
        return self._chan.is_closing()

    async def wait_closed(self) -> None:
        """Wait for the process to finish shutting down"""

        assert self._chan is not None
        await self._chan.wait_closed()

        for task in self._cleanup_tasks:
            await task

        self._cleanup_tasks = []


class SSHClientProcess(SSHProcess[AnyStr], SSHClientStreamSession[AnyStr]):
    """SSH client process handler"""

    _chan: SSHClientChannel[AnyStr]
    channel: SSHClientChannel[AnyStr]

    def __init__(self) -> None:
        super().__init__()

        self._stdin: Optional[SSHWriter[AnyStr]] = None
        self._stdout: Optional[SSHReader[AnyStr]] = None
        self._stderr: Optional[SSHReader[AnyStr]] = None

    def _collect_output(self, datatype: DataType = None) -> AnyStr:
        """Return output from the process"""

        recv_buf = self._recv_buf[datatype]

        if recv_buf and isinstance(recv_buf[-1], Exception):
            recv_buf, self._recv_buf[datatype] = recv_buf[:-1], recv_buf[-1:]
        else:
            self._recv_buf[datatype] = []

        buf = cast(AnyStr, '' if self._encoding else b'')
        return buf.join(cast(Iterable[AnyStr], recv_buf))

    def session_started(self) -> None:
        """Start a process for this newly opened client channel"""

        self._stdin = SSHWriter[AnyStr](self, self._chan)
        self._stdout = SSHReader[AnyStr](self, self._chan)
        self._stderr = SSHReader[AnyStr](self, self._chan, EXTENDED_DATA_STDERR)

    @property
    def exit_status(self) -> Optional[int]:
        """The exit status of the process"""

        return self._chan.get_exit_status()

    @property
    def exit_signal(self) -> Optional[Tuple[str, bool, str, str]]:
        """Exit signal information for the process"""

        return self._chan.get_exit_signal()

    @property
    def returncode(self) -> Optional[int]:
        """The exit status or negative exit signal number for the process"""

        return self._chan.get_returncode()

    @property
    def stdin(self) -> SSHWriter[AnyStr]:
        """The :class:`SSHWriter` to use to write to stdin of the process"""

        assert self._stdin is not None
        return self._stdin

    @property
    def stdout(self) -> SSHReader[AnyStr]:
        """The :class:`SSHReader` to use to read from stdout of the process"""

        assert self._stdout is not None
        return self._stdout

    @property
    def stderr(self) -> SSHReader[AnyStr]:
        """The :class:`SSHReader` to use to read from stderr of the process"""

        assert self._stderr is not None
        return self._stderr

    def feed_exception(self, exc: Exception) -> None:
        """Feed exception to the channel"""

        if isinstance(exc, TerminalSizeChanged):
            self._chan.change_terminal_size(exc.width, exc.height,
                                            exc.pixwidth, exc.pixheight)
        elif isinstance(exc, BreakReceived):
            self._chan.send_break(exc.msec)
        elif isinstance(exc, SignalReceived): # pragma: no branch
            self._chan.send_signal(exc.signal)

    async def redirect(self, stdin: Optional[ProcessSource] = None,
                       stdout: Optional[ProcessTarget] = None,
                       stderr: Optional[ProcessTarget] = None,
                       bufsize: int =io.DEFAULT_BUFFER_SIZE,
                       send_eof: bool = True, recv_eof: bool = True) -> None:
        """Perform I/O redirection for the process

           This method redirects data going to or from any or all of
           standard input, standard output, and standard error for
           the process.

           The `stdin` argument can be any of the following:

               * An :class:`SSHReader` object
               * An :class:`asyncio.StreamReader` object
               * A file object open for read
               * An `int` file descriptor open for read
               * A connected socket object
               * A string or :class:`PurePath <pathlib.PurePath>` containing
                 the name of a file or device to open
               * `DEVNULL` to provide no input to standard input
               * `PIPE` to interactively write standard input

           The `stdout` and `stderr` arguments can be any of the following:

               * An :class:`SSHWriter` object
               * An :class:`asyncio.StreamWriter` object
               * A file object open for write
               * An `int` file descriptor open for write
               * A connected socket object
               * A string or :class:`PurePath <pathlib.PurePath>` containing
                 the name of a file or device to open
               * `DEVNULL` to discard standard error output
               * `PIPE` to interactively read standard error output

           The `stderr` argument also accepts the value `STDOUT` to
           request that standard error output be delivered to stdout.

           File objects passed in can be associated with plain files, pipes,
           sockets, or ttys.

           The default value of `None` means to not change redirection
           for that stream.

           .. note:: While it is legal to use buffered I/O streams such
                     as sys.stdin, sys.stdout, and sys.stderr as redirect
                     targets, you must make sure buffers are flushed
                     before redirection begins and that these streams
                     are put back into blocking mode before attempting
                     to go back using buffered I/O again. Also, no buffered
                     I/O should be performed while redirection is active.

           .. note:: When passing in asyncio streams, it is the responsibility
                     of the caller to close the associated transport when it
                     is no longer needed.

           :param stdin:
               Source of data to feed to standard input
           :param stdout:
               Target to feed data from standard output to
           :param stderr:
               Target to feed data from standard error to
           :param bufsize:
               Buffer size to use when forwarding data from a file
           :param send_eof:
               Whether or not to send EOF to the channel when EOF is
               received from stdin, defaulting to `True`. If set to `False`,
               the channel will remain open after EOF is received on stdin,
               and multiple sources can be redirected to the channel.
           :param recv_eof:
               Whether or not to send EOF to stdout and stderr when EOF is
               received from the channel, defaulting to `True`. If set to
               `False`, the redirect targets of stdout and stderr will remain
               open after EOF is received on the channel and can be used for
               multiple redirects.
           :type bufsize: `int`
           :type send_eof: `bool`
           :type recv_eof: `bool`

        """

        if stdin:
            await self._create_reader(stdin, bufsize, send_eof, recv_eof)

        if stdout:
            await self._create_writer(stdout, bufsize, send_eof, recv_eof)

        if stderr:
            await self._create_writer(stderr, bufsize, send_eof, recv_eof,
                                      EXTENDED_DATA_STDERR)

    async def redirect_stdin(self, source: ProcessSource,
                             bufsize: int = io.DEFAULT_BUFFER_SIZE,
                             send_eof: bool = True) -> None:
        """Redirect standard input of the process"""

        await self.redirect(source, None, None, bufsize, send_eof, True)

    async def redirect_stdout(self, target: ProcessTarget,
                              bufsize: int = io.DEFAULT_BUFFER_SIZE,
                              recv_eof: bool = True) -> None:
        """Redirect standard output of the process"""

        await self.redirect(None, target, None, bufsize, True, recv_eof)

    async def redirect_stderr(self, target: ProcessTarget,
                              bufsize: int = io.DEFAULT_BUFFER_SIZE,
                              recv_eof: bool = True) -> None:
        """Redirect standard error of the process"""

        await self.redirect(None, None, target, bufsize, True, recv_eof)

    def collect_output(self) -> Tuple[AnyStr, AnyStr]:
        """Collect output from the process without blocking

           This method returns a tuple of the output that the process
           has written to stdout and stderr which has not yet been read.
           It is intended to be called instead of read() by callers
           that want to collect received data without blocking.

           :returns: A tuple of output to stdout and stderr

        """

        return (self._collect_output(),
                self._collect_output(EXTENDED_DATA_STDERR))

    # pylint: disable=redefined-builtin
    async def communicate(self, input: Optional[AnyStr] = None) -> \
            Tuple[AnyStr, AnyStr]:
        """Send input to and/or collect output from the process

           This method is a coroutine which optionally provides input
           to the process and then waits for the process to exit,
           returning a tuple of the data written to stdout and stderr.

           :param input:
               Input data to feed to standard input of the process. Data
               should be a `str` if encoding is set, or `bytes` if not.
           :type input: `str` or `bytes`

           :returns: A tuple of output to stdout and stderr

        """

        self._limit = 0
        self._maybe_resume_reading()

        if input:
            self._chan.write(input)
            self._chan.write_eof()

        await self.wait_closed()

        return self.collect_output()
    # pylint: enable=redefined-builtin

    def change_terminal_size(self, width: int, height: int,
                             pixwidth: int = 0, pixheight: int = 0) -> None:
        """Change the terminal window size for this process

           This method changes the width and height of the terminal
           associated with this process.

           :param width:
               The width of the terminal in characters
           :param height:
               The height of the terminal in characters
           :param pixwidth: (optional)
               The width of the terminal in pixels
           :param pixheight: (optional)
               The height of the terminal in pixels
           :type width: `int`
           :type height: `int`
           :type pixwidth: `int`
           :type pixheight: `int`

           :raises: :exc:`OSError` if the SSH channel is not open

        """

        self._chan.change_terminal_size(width, height, pixwidth, pixheight)

    def send_break(self, msec: int) -> None:
        """Send a break to the process

           :param msec:
               The duration of the break in milliseconds
           :type msec: `int`

           :raises: :exc:`OSError` if the SSH channel is not open

        """

        self._chan.send_break(msec)

    def send_signal(self, signal: str) -> None:
        """Send a signal to the process

           :param signal:
               The signal to deliver
           :type signal: `str`

           :raises: :exc:`OSError` if the SSH channel is not open

        """

        self._chan.send_signal(signal)

    def terminate(self) -> None:
        """Terminate the process

           :raises: :exc:`OSError` if the SSH channel is not open

        """

        self._chan.terminate()

    def kill(self) -> None:
        """Forcibly kill the process

           :raises: :exc:`OSError` if the SSH channel is not open

        """

        self._chan.kill()

    async def wait(self, check: bool = False,
                   timeout: Optional[float] = None) -> SSHCompletedProcess:
        """Wait for process to exit

           This method is a coroutine which waits for the process to
           exit. It returns an :class:`SSHCompletedProcess` object with
           the exit status or signal information and the output sent
           to stdout and stderr if those are redirected to pipes.

           If the check argument is set to `True`, a non-zero exit
           status from the process with trigger the :exc:`ProcessError`
           exception to be raised.

           If a timeout is specified and it expires before the process
           exits, the :exc:`TimeoutError` exception will be raised. By
           default, no timeout is set and this call will wait indefinitely.

           :param check:
               Whether or not to raise an error on non-zero exit status
           :param timeout:
               Amount of time in seconds to wait for process to exit, or
               `None` to wait indefinitely
           :type check: `bool`
           :type timeout: `int`, `float`, or `None`

           :returns: :class:`SSHCompletedProcess`

           :raises: | :exc:`ProcessError` if check is set to `True`
                      and the process returns a non-zero exit status
                    | :exc:`TimeoutError` if the timeout expires
                      before the process exits

        """

        try:
            stdout_data, stderr_data = \
                await asyncio.wait_for(self.communicate(), timeout)
        except asyncio.TimeoutError:
            stdout_data, stderr_data = self.collect_output()

            raise TimeoutError(self.env, self.command, self.subsystem,
                               self.exit_status, self.exit_signal,
                               self.returncode, stdout_data,
                               stderr_data) from None

        if check and self.exit_status:
            raise ProcessError(self.env, self.command, self.subsystem,
                               self.exit_status, self.exit_signal,
                               self.returncode, stdout_data, stderr_data)
        else:
            return SSHCompletedProcess(self.env, self.command, self.subsystem,
                                       self.exit_status, self.exit_signal,
                                       self.returncode, stdout_data,
                                       stderr_data)


class SSHServerProcess(SSHProcess[AnyStr], SSHServerStreamSession[AnyStr]):
    """SSH server process handler"""

    _chan: SSHServerChannel[AnyStr]
    channel: SSHServerChannel[AnyStr]

    def __init__(self, process_factory: SSHServerProcessFactory,
                 sftp_factory: Optional[SFTPServerFactory],
                 sftp_version: int, allow_scp: bool):
        super().__init__(self._start_process, sftp_factory,
                         sftp_version, allow_scp)

        self._process_factory = process_factory

        self._stdin: Optional[SSHReader[AnyStr]] = None
        self._stdout: Optional[SSHWriter[AnyStr]] = None
        self._stderr: Optional[SSHWriter[AnyStr]] = None

    def _start_process(self, stdin: SSHReader[AnyStr],
                       stdout: SSHWriter[AnyStr],
                       stderr: SSHWriter[AnyStr]) -> MaybeAwait[None]:
        """Start a new server process"""

        self._stdin = stdin
        self._stdout = stdout
        self._stderr = stderr

        return self._process_factory(self)

    @property
    def term_type(self) -> Optional[str]:
        """The terminal type set by the client

           If the client didn't request a pseudo-terminal, this
           property will be set to `None`.

        """

        return self._chan.get_terminal_type()

    @property
    def term_size(self) -> TermSize:
        """The terminal size set by the client

           This property contains a tuple of four `int` values
           representing the width and height of the terminal in
           characters followed by the width and height of the
           terminal in pixels. If the client hasn't set terminal
           size information, the values will be set to zero.

        """

        return self._chan.get_terminal_size()

    @property
    def term_modes(self) -> TermModes:
        """A mapping containing the TTY modes set by the client

           If the client didn't request a pseudo-terminal, this
           property will be set to an empty mapping.

        """

        return self._chan.get_terminal_modes()

    @property
    def stdin(self) -> SSHReader[AnyStr]:
        """The :class:`SSHReader` to use to read from stdin of the process"""

        assert self._stdin is not None
        return self._stdin

    @property
    def stdout(self) -> SSHWriter[AnyStr]:
        """The :class:`SSHWriter` to use to write to stdout of the process"""

        assert self._stdout is not None
        return self._stdout

    @property
    def stderr(self) -> SSHWriter[AnyStr]:
        """The :class:`SSHWriter` to use to write to stderr of the process"""

        assert self._stderr is not None
        return self._stderr

    def exception_received(self, exc: Exception) -> None:
        """Handle an incoming exception on the channel"""

        writer = self._writers.get(None)

        if writer:
            writer.write_exception(exc)
        else:
            super().exception_received(exc)

    async def redirect(self, stdin: Optional[ProcessTarget] = None,
                       stdout: Optional[ProcessSource] = None,
                       stderr: Optional[ProcessSource] = None,
                       bufsize: int = io.DEFAULT_BUFFER_SIZE,
                       send_eof: bool = True, recv_eof: bool = True) -> None:
        """Perform I/O redirection for the process

           This method redirects data going to or from any or all of
           standard input, standard output, and standard error for
           the process.

           The `stdin` argument can be any of the following:

               * An :class:`SSHWriter` object
               * An :class:`asyncio.StreamWriter` object
               * A file object open for write
               * An `int` file descriptor open for write
               * A connected socket object
               * A string or :class:`PurePath <pathlib.PurePath>` containing
                 the name of a file or device to open
               * `DEVNULL` to discard standard error output
               * `PIPE` to interactively read standard error output

           The `stdout` and `stderr` arguments can be any of the following:

               * An :class:`SSHReader` object
               * An :class:`asyncio.StreamReader` object
               * A file object open for read
               * An `int` file descriptor open for read
               * A connected socket object
               * A string or :class:`PurePath <pathlib.PurePath>` containing
                 the name of a file or device to open
               * `DEVNULL` to provide no input to standard input
               * `PIPE` to interactively write standard input

           File objects passed in can be associated with plain files, pipes,
           sockets, or ttys.

           The default value of `None` means to not change redirection
           for that stream.

           .. note:: When passing in asyncio streams, it is the responsibility
                     of the caller to close the associated transport when it
                     is no longer needed.

           :param stdin:
               Target to feed data from standard input to
           :param stdout:
               Source of data to feed to standard output
           :param stderr:
               Source of data to feed to standard error
           :param bufsize:
               Buffer size to use when forwarding data from a file
           :param send_eof:
               Whether or not to send EOF to the channel when EOF is
               received from stdout or stderr, defaulting to `True`. If
               set to `False`, the channel will remain open after EOF is
               received on stdout or stderr, and multiple sources can be
               redirected to the channel.
           :param recv_eof:
               Whether or not to send EOF to stdin when EOF is received
               on the channel, defaulting to `True`. If set to `False`,
               the redirect target of stdin will remain open after EOF
               is received on the channel and can be used for multiple
               redirects.
           :type bufsize: `int`
           :type send_eof: `bool`
           :type recv_eof: `bool`

        """

        if stdin:
            await self._create_writer(stdin, bufsize, send_eof, recv_eof)

        if stdout:
            await self._create_reader(stdout, bufsize, send_eof, recv_eof)

        if stderr:
            await self._create_reader(stderr, bufsize, send_eof, recv_eof,
                                      EXTENDED_DATA_STDERR)

    async def redirect_stdin(self, target: ProcessTarget,
                             bufsize: int = io.DEFAULT_BUFFER_SIZE,
                             recv_eof: bool = True) -> None:
        """Redirect standard input of the process"""

        await self.redirect(target, None, None, bufsize, True, recv_eof)

    async def redirect_stdout(self, source: ProcessSource,
                              bufsize: int = io.DEFAULT_BUFFER_SIZE,
                              send_eof: bool = True) -> None:
        """Redirect standard output of the process"""

        await self.redirect(None, source, None, bufsize, send_eof, True)

    async def redirect_stderr(self, source: ProcessSource,
                              bufsize: int = io.DEFAULT_BUFFER_SIZE,
                              send_eof: bool = True) -> None:
        """Redirect standard error of the process"""

        await self.redirect(None, None, source, bufsize, send_eof, True)

    def get_terminal_type(self) -> Optional[str]:
        """Return the terminal type set by the client for the process

           This method returns the terminal type set by the client
           when the process was started. If the client didn't request
           a pseudo-terminal, this method will return `None`.

           :returns: A `str` containing the terminal type or `None` if
                     no pseudo-terminal was requested

        """

        return self.term_type

    def get_terminal_size(self) -> Tuple[int, int, int, int]:
        """Return the terminal size set by the client for the process

           This method returns the latest terminal size information set
           by the client. If the client didn't set any terminal size
           information, all values returned will be zero.

           :returns: A tuple of four `int` values containing the width and
                     height of the terminal in characters and the width
                     and height of the terminal in pixels

        """

        return self.term_size

    def get_terminal_mode(self, mode: int) -> Optional[int]:
        """Return the requested TTY mode for this session

           This method looks up the value of a POSIX terminal mode
           set by the client when the process was started. If the client
           didn't request a pseudo-terminal or didn't set the requested
           TTY mode opcode, this method will return `None`.

           :param mode:
               POSIX terminal mode taken from :ref:`POSIX terminal modes
               <PTYModes>` to look up
           :type mode: `int`

           :returns: An `int` containing the value of the requested
                     POSIX terminal mode or `None` if the requested
                     mode was not set

        """

        return self.term_modes.get(mode)

    def exit(self, status: int) -> None:
        """Send exit status and close the channel

           This method can be called to report an exit status for the
           process back to the client and close the channel.

           :param status:
               The exit status to report to the client
           :type status: `int`

        """

        self._chan.exit(status)

    def exit_with_signal(self, signal: str, core_dumped: bool = False,
                         msg: str = '', lang: str = DEFAULT_LANG) -> None:
        """Send exit signal and close the channel

           This method can be called to report that the process
           terminated abnormslly with a signal. A more detailed
           error message may also provided, along with an indication
           of whether or not the process dumped core. After
           reporting the signal, the channel is closed.

           :param signal:
               The signal which caused the process to exit
           :param core_dumped: (optional)
               Whether or not the process dumped core
           :param msg: (optional)
               Details about what error occurred
           :param lang: (optional)
               The language the error message is in
           :type signal: `str`
           :type core_dumped: `bool`
           :type msg: `str`
           :type lang: `str`

        """

        return self._chan.exit_with_signal(signal, core_dumped, msg, lang)



# Copyright (c) 2013-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SSH key exchange handlers"""

import binascii
from hashlib import md5
from typing import TYPE_CHECKING, Dict, List, Sequence, Tuple, Type

from .logging import SSHLogger
from .misc import HashType
from .packet import SSHPacketHandler


if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .connection import SSHConnection


_KexAlgList = List[bytes]
_KexAlgMap = Dict[bytes, Tuple[Type['Kex'], HashType, Tuple]]


_kex_algs: _KexAlgList = []
_default_kex_algs:_KexAlgList = []
_kex_handlers: _KexAlgMap = {}

_gss_kex_algs: _KexAlgList = []
_default_gss_kex_algs: _KexAlgList = []
_gss_kex_handlers: _KexAlgMap = {}


class Kex(SSHPacketHandler):
    """Parent class for key exchange handlers"""

    def __init__(self, alg: bytes, conn: 'SSHConnection', hash_alg: HashType):
        self.algorithm = alg

        self._conn = conn
        self._logger = conn.logger
        self._hash_alg = hash_alg


    async def start(self) -> None:
        """Start key exchange"""

        raise NotImplementedError

    def send_packet(self, pkttype: int, *args: bytes) -> None:
        """Send a kex packet"""

        self._conn.send_packet(pkttype, *args, handler=self)

    @property
    def logger(self) -> SSHLogger:
        """A logger associated with this connection"""

        return self._logger

    def compute_key(self, k: bytes, h: bytes, x: bytes,
                    session_id: bytes, keylen: int) -> bytes:
        """Compute keys from output of key exchange"""

        key = b''
        while len(key) < keylen:
            hash_obj = self._hash_alg()
            hash_obj.update(k)
            hash_obj.update(h)
            hash_obj.update(key if key else x + session_id)
            key += hash_obj.digest()

        return key[:keylen]


def register_kex_alg(alg: bytes, handler: Type[Kex], hash_alg: HashType,
                     args: Tuple, default: bool) -> None:
    """Register a key exchange algorithm"""

    _kex_algs.append(alg)

    if default:
        _default_kex_algs.append(alg)

    _kex_handlers[alg] = (handler, hash_alg, args)


def register_gss_kex_alg(alg: bytes, handler: Type[Kex], hash_alg: HashType,
                         args: Tuple, default: bool) -> None:
    """Register a GSSAPI key exchange algorithm"""

    _gss_kex_algs.append(alg)

    if default:
        _default_gss_kex_algs.append(alg)

    _gss_kex_handlers[alg] = (handler, hash_alg, args)


def get_kex_algs() -> List[bytes]:
    """Return supported key exchange algorithms"""

    return _gss_kex_algs + _kex_algs


def get_default_kex_algs() -> List[bytes]:
    """Return default key exchange algorithms"""

    return _default_gss_kex_algs + _default_kex_algs


def expand_kex_algs(kex_algs: Sequence[bytes], mechs: Sequence[bytes],
                    host_key_available: bool) -> List[bytes]:
    """Add mechanisms to GSS entries in key exchange algorithm list"""

    expanded_kex_algs: List[bytes] = []

    for alg in kex_algs:
        if alg.startswith(b'gss-'):
            for mech in mechs:
                suffix = b'-' + binascii.b2a_base64(md5(mech).digest())[:-1]
                expanded_kex_algs.append(alg + suffix)
        elif host_key_available:
            expanded_kex_algs.append(alg)

    return expanded_kex_algs


def get_kex(conn: 'SSHConnection', alg: bytes) -> Kex:
    """Return a key exchange handler

       The function looks up a key exchange algorithm and returns a
       handler which can perform that type of key exchange.

    """

    if alg.startswith(b'gss-'):
        alg = alg.rsplit(b'-', 1)[0]
        handler, hash_alg, args = _gss_kex_handlers[alg]
    else:
        handler, hash_alg, args = _kex_handlers[alg]

    return handler(alg, conn, hash_alg, *args)



# Copyright (c) 2016-2023 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""Input line editor"""

import re

from functools import partial
from typing import TYPE_CHECKING, Callable, Dict, List
from typing import Optional, Set, Tuple, Union, cast
from unicodedata import east_asian_width

from .session import DataType


if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .channel import SSHServerChannel
    from .session import SSHServerSession


_CharDict = Dict[str, object]
_CharHandler = Callable[['SSHLineEditor'], None]
_KeyHandler = Callable[[str, int], Union[bool, Tuple[str, int]]]


_DEFAULT_WIDTH = 80

_ansi_terminals = ('ansi', 'cygwin', 'linux', 'putty', 'screen', 'teraterm',
                   'cit80', 'vt100', 'vt102', 'vt220', 'vt320', 'xterm',
                   'xterm-color', 'xterm-16color', 'xterm-256color', 'rxvt',
                   'rxvt-color')


def _is_wide(ch: str) -> bool:
    """Return display width of character"""

    return east_asian_width(ch) in 'WF'


class SSHLineEditor:
    """Input line editor"""

    def __init__(self, chan: 'SSHServerChannel[str]',
                 session: 'SSHServerSession[str]', line_echo: bool,
                 history_size: int, max_line_length: int, term_type: str,
                 width: int):
        self._chan = chan
        self._session = session
        self._line_echo = line_echo
        self._line_pending = False
        self._history_size = history_size if history_size > 0 else 0
        self._max_line_length = max_line_length
        self._wrap = term_type in _ansi_terminals
        self._width = width or _DEFAULT_WIDTH
        self._line_mode = True
        self._echo = True
        self._start_column = 0
        self._end_column = 0
        self._cursor = 0
        self._left_pos = 0
        self._right_pos = 0
        self._pos = 0
        self._line = ''
        self._bell_rung = False
        self._early_wrap: Set[int] = set()
        self._outbuf: List[str] = []
        self._keymap: _CharDict = {}
        self._key_state = self._keymap
        self._erased = ''
        self._history: List[str] = []
        self._history_index = 0

        for func, keys in self._keylist:
            for key in keys:
                self._add_key(key, func)

        self._build_printable()

    def _add_key(self, key: str, func: _CharHandler) -> None:
        """Add a key to the keymap"""

        keymap = self._keymap

        for ch in key[:-1]:
            if ch not in keymap:
                keymap[ch] = {}

            keymap = cast(_CharDict, keymap[ch])

        keymap[key[-1]] = func

    def _del_key(self, key: str) -> None:
        """Delete a key from the keymap"""

        keymap = self._keymap

        for ch in key[:-1]:
            if ch not in keymap:
                return

            keymap = cast(_CharDict, keymap[ch])

        keymap.pop(key[-1], None)

    def _build_printable(self) -> None:
        """Build a regex of printable ASCII non-registered keys"""

        def _escape(c: int) -> str:
            """Backslash escape special characters in regex character range"""

            ch = chr(c)
            return ('\\' if (ch in '-&|[]\\^~') else '') + ch

        def _is_printable(ch: str) -> bool:
            """Return if character is printable and has no handler"""

            return ch.isprintable() and ch not in keys

        pat: List[str] = []
        keys = self._keymap.keys()
        start = ord(' ')
        limit = 0x10000

        while start < limit:
            while start < limit and not _is_printable(chr(start)):
                start += 1

            end = start

            while _is_printable(chr(end)):
                end += 1

            pat.append(_escape(start))

            if start != end - 1:
                pat.append('-' + _escape(end - 1))

            start = end + 1

        self._printable = re.compile('[' + ''.join(pat) + ']*')

    def _char_width(self, pos: int) -> int:
        """Return width of character at specified position"""

        return 1 + _is_wide(self._line[pos]) + ((pos + 1) in self._early_wrap)

    def _determine_column(self, data: str, column: int,
                          pos: Optional[int] = None) -> Tuple[str, int]:
        """Determine new output column after output occurs"""

        escaped = False
        offset = pos
        last_wrap_pos = pos
        wrapped_data = []

        for ch in data:
            if ch == '\b':
                column -= 1
            elif ch == '\x1b':
                escaped = True
            elif escaped:
                if ch == 'm':
                    escaped = False
            else:
                if _is_wide(ch) and (column % self._width) == self._width - 1:
                    column += 1

                    if pos is not None:
                        assert last_wrap_pos is not None
                        assert offset is not None

                        wrapped_data.append(data[last_wrap_pos - offset:
                                                 pos - offset])
                        last_wrap_pos = pos

                        self._early_wrap.add(pos)
                else:
                    if pos is not None:
                        self._early_wrap.discard(pos)

                column += 1 + _is_wide(ch)

            if pos is not None:
                pos += 1

        if pos is not None:
            assert last_wrap_pos is not None
            assert offset is not None

            wrapped_data.append(data[last_wrap_pos - offset:])
            return ' '.join(wrapped_data), column
        else:
            return data, column

    def _output(self, data: str, pos: Optional[int] = None) -> None:
        """Generate output and calculate new output column"""

        idx = data.rfind('\n')

        if idx >= 0:
            self._outbuf.append(data[:idx+1])
            tail = data[idx+1:]
            self._cursor = 0
        else:
            tail = data

        data, self._cursor = self._determine_column(tail, self._cursor, pos)

        self._outbuf.append(data)

        if self._cursor and self._cursor % self._width == 0:
            self._outbuf.append(' \b')

    def _ring_bell(self) -> None:
        """Ring the terminal bell"""

        if not self._bell_rung:
            self._outbuf.append('\a')
            self._bell_rung = True

    def _update_input_window(self, new_pos: int) -> int:
        """Update visible input window when not wrapping onto multiple lines"""

        line_len = len(self._line)

        if new_pos < self._left_pos:
            self._left_pos = new_pos
        else:
            if new_pos < line_len:
                new_pos += 1

            pos = self._pos
            column = self._cursor

            while pos < new_pos:
                column += self._char_width(pos)
                pos += 1

            if column >= self._width:
                while column >= self._width:
                    column -= self._char_width(self._left_pos)
                    self._left_pos += 1
            else:
                while self._left_pos > 0:
                    column += self._char_width(self._left_pos)

                    if column < self._width:
                        self._left_pos -= 1
                    else:
                        break

        column = self._start_column
        self._right_pos = self._left_pos

        while self._right_pos < line_len:
            ch_width = self._char_width(self._right_pos)

            if column + ch_width < self._width:
                self._right_pos += 1
                column += ch_width
            else:
                break

        return column

    def _move_cursor(self, column: int) -> None:
        """Move the cursor to selected position in input line"""

        start_row = self._cursor // self._width
        start_col = self._cursor % self._width

        end_row = column // self._width
        end_col = column % self._width

        if end_row < start_row:
            self._outbuf.append('\x1b[' + str(start_row-end_row) + 'A')
        elif end_row > start_row:
            self._outbuf.append('\x1b[' + str(end_row-start_row) + 'B')

        if end_col > start_col:
            self._outbuf.append('\x1b[' + str(end_col-start_col) + 'C')
        elif end_col < start_col:
            self._outbuf.append('\x1b[' + str(start_col-end_col) + 'D')

        self._cursor = column

    def _move_back(self, column: int) -> None:
        """Move the cursor backward to selected position in input line"""

        if self._wrap:
            self._move_cursor(column)
        else:
            self._outbuf.append('\b' * (self._cursor - column))
            self._cursor = column

    def _clear_to_end(self) -> None:
        """Clear any remaining characters from previous input line"""

        column = self._cursor
        remaining = self._end_column - column

        if remaining > 0:
            self._outbuf.append(' ' * remaining)
            self._cursor = self._end_column

            if self._cursor % self._width == 0:
                self._outbuf.append(' \b')

        self._move_back(column)
        self._end_column = column

    def _erase_input(self) -> None:
        """Erase current input line"""

        self._move_cursor(self._start_column)
        self._clear_to_end()
        self._early_wrap.clear()

    def _draw_input(self) -> None:
        """Draw current input line"""

        if self._line and self._echo:
            if self._wrap:
                self._output(self._line[:self._pos], 0)
                column = self._cursor
                self._output(self._line[self._pos:], self._pos)
            else:
                self._update_input_window(self._pos)
                self._output(self._line[self._left_pos:self._pos])
                column = self._cursor
                self._output(self._line[self._pos:self._right_pos])

            self._end_column = self._cursor
            self._move_back(column)

    def _reposition(self, new_pos: int, new_column: int) -> None:
        """Reposition the cursor to selected position in input"""

        if self._echo:
            if self._wrap:
                self._move_cursor(new_column)
            else:
                self._update_input(self._pos, self._cursor, new_pos)

        self._pos = new_pos

    def _update_input(self, pos: int, column: int, new_pos: int) -> None:
        """Update selected portion of current input line"""

        if self._echo:
            if self._wrap:
                if pos in self._early_wrap:
                    column -= 1

                self._move_cursor(column)
                prev_wrap = new_pos in self._early_wrap
                self._output(self._line[pos:new_pos], pos)
                column = self._cursor
                self._output(self._line[new_pos:], new_pos)
                column += (new_pos in self._early_wrap) - prev_wrap
            else:
                self._update_input_window(new_pos)
                self._move_back(self._start_column)
                self._output(self._line[self._left_pos:new_pos])
                column = self._cursor
                self._output(self._line[new_pos:self._right_pos])

            self._clear_to_end()
            self._move_back(column)

        self._pos = new_pos

    def _reset_line(self) -> None:
        """Reset input line to empty"""

        self._line = ''
        self._left_pos = 0
        self._right_pos = 0
        self._pos = 0
        self._start_column = self._cursor
        self._end_column = self._cursor

    def _reset_pending(self) -> None:
        """Reset a pending echoed line if any"""

        if self._line_pending:
            self._erase_input()
            self._reset_line()
            self._line_pending = False

    def _insert_printable(self, data: str) -> None:
        """Insert data into the input line"""

        line_len = len(self._line)
        data_len = len(data)

        if self._max_line_length:
            if line_len + data_len > self._max_line_length:
                self._ring_bell()
                data_len = self._max_line_length - line_len
                data = data[:data_len]

        if data:
            pos = self._pos
            new_pos = pos + data_len
            self._line = self._line[:pos] + data + self._line[pos:]

            self._update_input(pos, self._cursor, new_pos)

    def _end_line(self) -> None:
        """End the current input line and send it to the session"""

        line = self._line

        need_wrap = (self._echo and not self._wrap and
                     (self._left_pos > 0 or self._right_pos < len(line)))

        if self._line_echo or need_wrap:
            if need_wrap:
                self._output('\b' * (self._cursor - self._start_column) + line)
            else:
                self._move_to_end()

            self._output('\r\n')
            self._reset_line()
        else:
            self._move_to_end()
            self._line_pending = True

        if self._echo and self._history_size and line:
            self._history.append(line)
            self._history = self._history[-self._history_size:]

        self._history_index = len(self._history)

        self._session.data_received(line + '\n', None)

    def _eof_or_delete(self) -> None:
        """Erase character to the right, or send EOF if input line is empty"""

        if not self._line:
            self._session.soft_eof_received()
        else:
            self._erase_right()

    def _erase_left(self) -> None:
        """Erase character to the left"""

        if self._pos > 0:
            pos = self._pos - 1
            column = self._cursor - self._char_width(pos)
            self._line = self._line[:pos] + self._line[pos+1:]
            self._update_input(pos, column, pos)
        else:
            self._ring_bell()

    def _erase_right(self) -> None:
        """Erase character to the right"""

        if self._pos < len(self._line):
            pos = self._pos
            self._line = self._line[:pos] + self._line[pos+1:]
            self._update_input(pos, self._cursor, pos)
        else:
            self._ring_bell()

    def _erase_line(self) -> None:
        """Erase entire input line"""

        self._erased = self._line
        self._line = ''
        self._update_input(0, self._start_column, 0)

    def _erase_to_end(self) -> None:
        """Erase to end of input line"""

        pos = self._pos
        self._erased = self._line[pos:]
        self._line = self._line[:pos]
        self._update_input(pos, self._cursor, pos)

    def _handle_key(self, key: str, handler: _KeyHandler) -> None:
        """Call an external key handler"""

        result = handler(self._line, self._pos)

        if result is True:
            if key.isprintable():
                self._insert_printable(key)
            else:
                self._ring_bell()
        elif result is False:
            self._ring_bell()
        else:
            line, new_pos = cast(Tuple[str, int], result)

            if new_pos < 0:
                self._session.signal_received(line)
            else:
                self._line = line
                self._update_input(0, self._start_column, new_pos)

    def _history_prev(self) -> None:
        """Replace input with previous line in history"""

        if self._history_index > 0:
            self._history_index -= 1
            self._line = self._history[self._history_index]
            self._update_input(0, self._start_column, len(self._line))
        else:
            self._ring_bell()

    def _history_next(self) -> None:
        """Replace input with next line in history"""

        if self._history_index < len(self._history):
            self._history_index += 1

            if self._history_index < len(self._history):
                self._line = self._history[self._history_index]
            else:
                self._line = ''

            self._update_input(0, self._start_column, len(self._line))
        else:
            self._ring_bell()

    def _move_left(self) -> None:
        """Move left in input line"""

        if self._pos > 0:
            pos = self._pos - 1
            column = self._cursor - self._char_width(pos)
            self._reposition(pos, column)
        else:
            self._ring_bell()

    def _move_right(self) -> None:
        """Move right in input line"""

        if self._pos < len(self._line):
            pos = self._pos
            column = self._cursor + self._char_width(pos)
            self._reposition(pos + 1, column)
        else:
            self._ring_bell()

    def _move_to_start(self) -> None:
        """Move to start of input line"""

        self._reposition(0, self._start_column)

    def _move_to_end(self) -> None:
        """Move to end of input line"""

        self._reposition(len(self._line), self._end_column)

    def _redraw(self) -> None:
        """Redraw input line"""

        self._erase_input()
        self._draw_input()

    def _insert_erased(self) -> None:
        """Insert previously erased input"""

        self._insert_printable(self._erased)

    def _send_break(self) -> None:
        """Send break to session"""

        self._session.break_received(0)

    _keylist = ((_end_line,      ('\n', '\r', '\x1bOM')),
                (_eof_or_delete, ('\x04',)),
                (_erase_left,    ('\x08', '\x7f')),
                (_erase_right,   ('\x1b[3~',)),
                (_erase_line,    ('\x15',)),
                (_erase_to_end,  ('\x0b',)),
                (_history_prev,  ('\x10', '\x1b[A', '\x1bOA')),
                (_history_next,  ('\x0e', '\x1b[B', '\x1bOB')),
                (_move_left,     ('\x02', '\x1b[D', '\x1bOD')),
                (_move_right,    ('\x06', '\x1b[C', '\x1bOC')),
                (_move_to_start, ('\x01', '\x1b[H', '\x1b[1~')),
                (_move_to_end,   ('\x05', '\x1b[F', '\x1b[4~')),
                (_redraw,        ('\x12',)),
                (_insert_erased, ('\x19',)),
                (_send_break,    ('\x03', '\x1b[33~')))

    def register_key(self, key: str, handler: _KeyHandler) -> None:
        """Register a handler to be called when a key is pressed"""

        self._add_key(key, partial(SSHLineEditor._handle_key,
                                   key=key, handler=handler))
        self._build_printable()

    def unregister_key(self, key: str) -> None:
        """Remove the handler associated with a key"""

        self._del_key(key)
        self._build_printable()

    def set_input(self, line: str, pos: int) -> None:
        """Set input line and cursor position"""

        self._reset_pending()

        self._line = line
        self._update_input(0, self._start_column, pos)

    def set_line_mode(self, line_mode: bool) -> None:
        """Enable/disable input line editing"""

        self._reset_pending()

        if self._line and not line_mode:
            data = self._line
            self._erase_input()
            self._line = ''

            self._session.data_received(data, None)

        self._line_mode = line_mode

    def set_echo(self, echo: bool) -> None:
        """Enable/disable echoing of input in line mode"""

        self._reset_pending()

        if self._echo and not echo:
            self._erase_input()
            self._echo = False
        elif echo and not self._echo:
            self._echo = True
            self._draw_input()

    def set_width(self, width: int) -> None:
        """Set terminal line width"""

        self._reset_pending()

        self._width = width or _DEFAULT_WIDTH

        if self._wrap:
            _, self._cursor = self._determine_column(self._line,
                                                     self._start_column, 0)

        self._redraw()

    def process_input(self, data: str, datatype: DataType) -> None:
        """Process input from channel"""

        if self._line_mode:
            data_len = len(data)
            idx = 0

            while idx < data_len:
                self._reset_pending()

                ch = data[idx]
                idx += 1

                if ch in self._key_state:
                    key_state = self._key_state[ch]

                    if callable(key_state):
                        try:
                            cast(_CharHandler, key_state)(self)
                        finally:
                            self._key_state = self._keymap
                    else:
                        self._key_state = cast(_CharDict, key_state)
                elif self._key_state == self._keymap and ch.isprintable():
                    match = self._printable.match(data, idx - 1)

                    assert match is not None
                    match = match[0]

                    if match:
                        self._insert_printable(match)
                        idx += len(match) - 1
                    else:
                        self._insert_printable(ch)
                else:
                    self._key_state = self._keymap
                    self._ring_bell()

            self._bell_rung = False

            if self._outbuf:
                self._chan.write(''.join(self._outbuf))
                self._outbuf.clear()
        else:
            self._session.data_received(data, datatype)

    def process_output(self, data: str) -> None:
        """Process output to channel"""

        if self._line_pending:
            if data.startswith(self._line):
                self._start_column = self._cursor
                data = data[len(self._line):]
            else:
                self._erase_input()

            self._reset_line()
            self._line_pending = False

        data = data.replace('\n', '\r\n')

        self._erase_input()
        self._output(data)

        if not self._wrap:
            self._cursor %= self._width

        self._start_column = self._cursor
        self._end_column = self._cursor
        self._draw_input()

        self._chan.write(''.join(self._outbuf))
        self._outbuf.clear()


class SSHLineEditorChannel:
    """Input line editor channel wrapper

       When creating server channels with `line_editor` set to `True`,
       this class is wrapped around the channel, providing the caller with
       the ability to enable and disable input line editing and echoing.

       .. note:: Line editing is only available when a pseudo-terminal
                 is requested on the server channel and the character
                 encoding on the channel is not set to `None`.

    """

    def __init__(self, orig_chan: 'SSHServerChannel[str]',
                 orig_session: 'SSHServerSession[str]', line_echo: bool,
                 history_size: int, max_line_length: int):
        self._orig_chan = orig_chan
        self._orig_session = orig_session
        self._line_echo = line_echo
        self._history_size = history_size
        self._max_line_length = max_line_length
        self._editor: Optional[SSHLineEditor] = None

    def __getattr__(self, attr: str):
        """Delegate most channel functions to original channel"""

        return getattr(self._orig_chan, attr)

    def create_editor(self) -> Optional[SSHLineEditor]:
        """Create input line editor if encoding and terminal type are set"""

        encoding, _ = self._orig_chan.get_encoding()
        term_type = self._orig_chan.get_terminal_type()
        width = self._orig_chan.get_terminal_size()[0]

        if encoding and term_type:
            self._editor = SSHLineEditor(
                self._orig_chan, self._orig_session, self._line_echo,
                self._history_size, self._max_line_length, term_type, width)

        return self._editor

    def register_key(self, key: str, handler: _KeyHandler) -> None:
        """Register a handler to be called when a key is pressed

           This method registers a handler function which will be called
           when a user presses the specified key while inputting a line.

           The handler will be called with arguments of the current
           input line and cursor position, and updated versions of these
           two values should be returned as a tuple.

           The handler can also return a tuple of a signal name and
           negative cursor position to cause a signal to be delivered
           on the channel. In this case, the current input line is left
           unchanged but the signal is delivered before processing any
           additional input. This can be used to define "hot keys" that
           trigger actions unrelated to editing the input.

           If the registered key is printable text, returning `True` will
           insert that text at the current cursor position, acting as if
           no handler was registered for that key. This is useful if you
           want to perform a special action in some cases but not others,
           such as based on the current cursor position.

           Returning `False` will ring the bell and leave the input
           unchanged, indicating the requested action could not be
           performed.

           :param key:
               The key sequence to look for
           :param handler:
               The handler function to call when the key is pressed
           :type key: `str`
           :type handler: `callable`

        """

        assert self._editor is not None
        self._editor.register_key(key, handler)

    def unregister_key(self, key: str) -> None:
        """Remove the handler associated with a key

           This method removes a handler function associated with
           the specified key. If the key sequence is printable,
           this will cause it to return to being inserted at the
           current position when pressed. Otherwise, it will cause
           the bell to ring to signal the key is not understood.

           :param key:
               The key sequence to look for
           :type key: `str`

        """

        assert self._editor is not None
        self._editor.unregister_key(key)

    def clear_input(self) -> None:
        """Clear input line

           This method clears the current input line.

        """

        assert self._editor is not None
        self._editor.set_input('', 0)

    def set_input(self, line: str, pos: int) -> None:
        """Clear input line

           This method sets the current input line and cursor position.

           :param line:
               The new input line
           :param pos:
               The new cursor position within the input line
           :type line: `str`
           :type pos: `int`

        """

        assert self._editor is not None
        self._editor.set_input(line, pos)

    def set_line_mode(self, line_mode: bool) -> None:
        """Enable/disable input line editing

           This method enabled or disables input line editing. When set,
           only full lines of input are sent to the session, and each
           line of input can be edited before it is sent.

           :param line_mode:
               Whether or not to process input a line at a time
           :type line_mode: `bool`

        """

        self._orig_chan.logger.info('%s line editor',
                                    'Enabling' if line_mode else 'Disabling')

        assert self._editor is not None
        self._editor.set_line_mode(line_mode)

    def set_echo(self, echo: bool) -> None:
        """Enable/disable echoing of input in line mode

           This method enables or disables echoing of input data when
           input line editing is enabled.

           :param echo:
               Whether or not input to echo input as it is entered
           :type echo: `bool`

        """

        self._orig_chan.logger.info('%s echo',
                                    'Enabling' if echo else 'Disabling')

        assert self._editor is not None
        self._editor.set_echo(echo)

    def write(self, data: str, datatype: DataType = None) -> None:
        """Process data written to the channel"""

        if self._editor and datatype is None:
            self._editor.process_output(data)
        else:
            self._orig_chan.write(data, datatype)


class SSHLineEditorSession:
    """Input line editor session wrapper"""

    def __init__(self, chan: SSHLineEditorChannel,
                 orig_session: 'SSHServerSession[str]'):
        self._chan = chan
        self._orig_session = orig_session
        self._editor: Optional[SSHLineEditor] = None

    def __getattr__(self, attr: str):
        """Delegate most channel functions to original session"""

        return getattr(self._orig_session, attr)

    def session_started(self) -> None:
        """Start a session for this newly opened server channel"""

        self._editor = self._chan.create_editor()
        self._orig_session.session_started()

    def terminal_size_changed(self, width: int, height: int,
                              pixwidth: int, pixheight: int) -> None:
        """The terminal size has changed"""

        if self._editor:
            self._editor.set_width(width)

        self._orig_session.terminal_size_changed(width, height,
                                                 pixwidth, pixheight)

    def data_received(self, data: str, datatype: DataType) -> None:
        """Process data received from the channel"""

        if self._editor:
            self._editor.process_input(data, datatype)
        else:
            self._orig_session.data_received(data, datatype)

    def eof_received(self) -> Optional[bool]:
        """Process EOF received from the channel"""

        if self._editor:
            self._editor.set_line_mode(False)

        return self._orig_session.eof_received()



# Copyright (c) 2015-2021 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""Chacha20-Poly1305 symmetric encryption handler"""

from typing import Optional, Tuple

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends.openssl import backend
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import ChaCha20
from cryptography.hazmat.primitives.poly1305 import Poly1305

from .cipher import register_cipher


chacha_available = backend.poly1305_supported()


_CTR_0 = (0).to_bytes(8, 'little')
_CTR_1 = (1).to_bytes(8, 'little')

_POLY1305_KEYBYTES = 32

def chacha20(key: bytes, data: bytes, nonce: bytes, ctr: int) -> bytes:
    """Encrypt/decrypt a block of data with the ChaCha20 cipher"""

    return Cipher(ChaCha20(key, (_CTR_1 if ctr else _CTR_0) + nonce),
                  mode=None).encryptor().update(data)

def poly1305_key(key: bytes, nonce: bytes) -> bytes:
    """Derive a Poly1305 key"""

    return chacha20(key, _POLY1305_KEYBYTES * b'\0', nonce, 0)

def poly1305(key: bytes, data: bytes, nonce: bytes) -> bytes:
    """Compute a Poly1305 tag for a block of data"""

    return Poly1305.generate_tag(poly1305_key(key, nonce), data)

def poly1305_verify(key: bytes, data: bytes,
                    nonce: bytes, tag: bytes) -> bool:
    """Verify a Poly1305 tag for a block of data"""

    try:
        Poly1305.verify_tag(poly1305_key(key, nonce), data, tag)
        return True
    except InvalidSignature:
        return False


class ChachaCipher:
    """Shim for Chacha20-Poly1305 symmetric encryption"""

    def __init__(self, key: bytes):
        keylen = len(key) // 2
        self._key = key[:keylen]
        self._adkey = key[keylen:]

    def encrypt_and_sign(self, header: bytes, data: bytes,
                         nonce: bytes) -> Tuple[bytes, bytes]:
        """Encrypt and sign a block of data"""

        header = chacha20(self._adkey, header, nonce, 0)
        data = chacha20(self._key, data, nonce, 1)
        tag = poly1305(self._key, header + data, nonce)

        return header + data, tag

    def decrypt_header(self, header: bytes, nonce: bytes) -> bytes:
        """Decrypt header data"""

        return chacha20(self._adkey, header, nonce, 0)

    def verify_and_decrypt(self, header: bytes, data: bytes,
                           nonce: bytes, tag: bytes) -> Optional[bytes]:
        """Verify the signature of and decrypt a block of data"""

        if poly1305_verify(self._key, header + data, nonce, tag):
            return chacha20(self._key, data, nonce, 1)
        else:
            return None


if chacha_available: # pragma: no branch
    register_cipher('chacha20-poly1305', 64, 0, 1)



# Copyright (c) 2019-2023 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SSH subprocess handlers"""

from typing import TYPE_CHECKING, Any, AnyStr, Callable
from typing import Dict, Generic, Iterable, Optional

from .constants import EXTENDED_DATA_STDERR
from .process import SSHClientProcess
from .session import DataType


if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .channel import SSHChannel, SSHClientChannel


SubprocessFactory = Callable[[], 'SSHSubprocessProtocol']


class SSHSubprocessPipe(Generic[AnyStr]):
    """SSH subprocess pipe"""

    def __init__(self, chan: 'SSHClientChannel[AnyStr]',
                 datatype: DataType = None):
        self._chan: 'SSHClientChannel[AnyStr]' = chan
        self._datatype = datatype

    def close(self) -> None:
        """Shut down the remote process"""

        self._chan.close()

    def get_extra_info(self, name: str, default: Any = None) -> Any:
        """Return additional information about the remote process

           This method returns extra information about the channel
           associated with this subprocess. See :meth:`get_extra_info()
           <SSHClientChannel.get_extra_info>` on :class:`SSHClientChannel`
           for additional information.

        """

        return self._chan.get_extra_info(name, default)


class SSHSubprocessReadPipe(SSHSubprocessPipe[AnyStr]):
    """SSH subprocess pipe reader"""

    def pause_reading(self) -> None:
        """Pause delivery of incoming data from the remote process"""

        self._chan.pause_reading()

    def resume_reading(self) -> None:
        """Resume delivery of incoming data from the remote process"""

        self._chan.resume_reading()


class SSHSubprocessWritePipe(SSHSubprocessPipe[AnyStr]):
    """SSH subprocess pipe writer"""

    def abort(self) -> None:
        """Forcibly close the channel to the remote process"""

        self._chan.abort()

    def can_write_eof(self) -> bool:
        """Return whether the pipe supports :meth:`write_eof`"""

        return self._chan.can_write_eof()

    def get_write_buffer_size(self) -> int:
        """Return the current size of the pipe's output buffer"""

        return self._chan.get_write_buffer_size()

    def set_write_buffer_limits(self, high: Optional[int] = None,
                                low: Optional[int] = None) -> None:
        """Set the high- and low-water limits for write flow control"""

        self._chan.set_write_buffer_limits(high, low)

    def write(self, data: AnyStr) -> None:
        """Write data on this pipe"""

        self._chan.write(data, self._datatype)

    def writelines(self, list_of_data: Iterable[AnyStr]) -> None:
        """Write a list of data bytes on this pipe"""

        self._chan.writelines(list_of_data, self._datatype)

    def write_eof(self) -> None:
        """Write EOF on this pipe"""

        self._chan.write_eof()


class SSHSubprocessProtocol(Generic[AnyStr]):
    """SSH subprocess protocol

       This class conforms to :class:`asyncio.SubprocessProtocol`, but with
       the following enhancement:

           * If encoding is set when the subprocess is created, all data
             passed to :meth:`pipe_data_received` will be string values
             containing Unicode data. However, for compatibility with
             :class:`asyncio.SubprocessProtocol`, encoding defaults to
             `None`, in which case all data is delivered as bytes.

    """

    def connection_made(self,
                        transport: 'SSHSubprocessTransport[AnyStr]') -> None:
        """Called when a remote process is successfully started

           This method is called when a remote process is successfully
           started. The transport parameter should be stored if needed
           for later use.

           :param transport:
               The transport to use to communicate with the remote process.
           :type transport: :class:`SSHSubprocessTransport`

        """

    def pipe_data_received(self, fd: int, data: AnyStr) -> None:
        """Called when data is received from the remote process

           This method is called when data is received from the remote
           process. If an encoding was specified when the process was
           started, the data will be delivered as a string after decoding
           with the requested encoding. Otherwise, the data will be
           delivered as bytes.

           :param fd:
               The integer file descriptor of the pipe data was received
               on. This will be 1 for stdout or 2 for stderr.
           :param data:
               The data received from the remote process
           :type fd: `int`
           :type data: `str` or `bytes`

        """

    def pipe_connection_lost(self, fd: int, exc: Optional[Exception]) -> None:
        """Called when the pipe to a remote process is closed

           This method is called when a pipe to a remote process is
           closed. If the channel is shut down cleanly, *exc* will be
           `None`. Otherwise, it will be an exception explaining the
           reason the pipe was closed.

           :param fd:
               The integer file descriptor of the pipe which was
               closed. This will be 1 for stdout or 2 for stderr.
           :param exc:
               The exception which caused the channel to close, or
               `None` if the channel closed cleanly.
           :type fd: `int`
           :type exc: :class:`Exception` or `None`

        """

    def process_exited(self) -> None:
        """Called when a remote process has exited

           This method is called when the remote process has exited.
           Exit status information can be retrieved by calling
           :meth:`get_returncode() <SSHSubprocessTransport.get_returncode>`
           on the transport provided in :meth:`connection_made`.

        """


class SSHSubprocessTransport(SSHClientProcess[AnyStr]):
    """SSH subprocess transport

       This class conforms to :class:`asyncio.SubprocessTransport`, but with
       the following enhancements:

           * All functionality available through :class:`SSHClientProcess`
             is also available here, such as the ability to dynamically
             redirect stdin, stdout, and stderr at any time during the
             lifetime of the process.

           * If encoding is set when the subprocess is created, all data
             written to the transports created by :meth:`get_pipe_transport`
             should be strings containing Unicode data. The encoding defaults
             to `None`, though, to preserve compatibility with
             :class:`asyncio.SubprocessTransport`, which expects data
             to be written as bytes.

    """

    _chan: 'SSHClientChannel[AnyStr]'

    def __init__(self, protocol_factory: SubprocessFactory):
        super().__init__()

        self._pipes: Dict[int, SSHSubprocessPipe[AnyStr]] = {}
        self._protocol: SSHSubprocessProtocol[AnyStr] = protocol_factory()

    def get_protocol(self) -> SSHSubprocessProtocol[AnyStr]:
        """Return the subprocess protocol associated with this transport"""

        return self._protocol

    def connection_made(self, chan: 'SSHChannel[AnyStr]') -> None:
        """Handle a newly opened channel"""

        super().connection_made(chan)

        self._protocol.connection_made(self)

        self._pipes = {
            0: SSHSubprocessWritePipe(self._chan),
            1: SSHSubprocessReadPipe(self._chan),
            2: SSHSubprocessReadPipe(self._chan, EXTENDED_DATA_STDERR)
        }

    def session_started(self) -> None:
        """Override SSHClientProcess to avoid creating SSHReader/SSHWriter
           streams, since this class uses read/write pipe objects instead"""

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Handle an incoming channel close"""

        self._protocol.pipe_connection_lost(1, exc)
        self._protocol.pipe_connection_lost(2, exc)
        super().connection_lost(exc)

    def data_received(self, data: AnyStr, datatype: DataType) -> None:
        """Handle incoming data from the remote process"""

        writer = self._writers.get(datatype)

        if writer:
            writer.write(data)
        else:
            fd = 2 if datatype == EXTENDED_DATA_STDERR else 1
            self._protocol.pipe_data_received(fd, data)

    def exit_status_received(self, status: int) -> None:
        """Handle exit status for the remote process"""

        super().exit_status_received(status)
        self._protocol.process_exited()

    def exit_signal_received(self, signal: str, core_dumped: bool,
                             msg: str, lang: str) -> None:
        """Handle exit signal for the remote process"""

        super().exit_signal_received(signal, core_dumped, msg, lang)
        self._protocol.process_exited()

    def get_pid(self) -> Optional[int]:
        """Return the PID of the remote process

           This method always returns `None`, since SSH doesn't report
           remote PIDs.

        """

        # pylint: disable=no-self-use

        return None

    def get_pipe_transport(self, fd: int) -> \
            Optional[SSHSubprocessPipe[AnyStr]]:
        """Return a transport for the requested stream

           :param fd:
               The integer file descriptor (0-2) to return the transport for,
               where 0 means stdin, 1 means stdout, and 2 means stderr.
           :type fd: `int`

           :returns: an :class:`SSHSubprocessReadPipe` or
                     :class:`SSHSubprocessWritePipe`

        """

        return self._pipes.get(fd)

    def get_returncode(self) -> Optional[int]:
        """Return the exit status or signal for the remote process

           This method returns the exit status of the session if one has
           been sent. If an exit signal was sent, this method returns
           the negative of the numeric value of that signal, matching
           the behavior of :meth:`asyncio.SubprocessTransport.get_returncode`.
           If neither has been sent, this method returns `None`.

           :returns: `int` or `None`

        """

        return self.returncode
